# -------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Aug 19, 2009 by: rch

from math import \
    cos, sin

from ibvpy.tmodel.mats2D.mats2D_eval import MATS2DEval
from ibvpy.tmodel.mats2D.mats2D_tensor import \
    map2d_tns2_to_tns4, \
    get_D_plane_stress, get_D_plane_strain, get_C_plane_stress, get_C_plane_strain
from ibvpy.tmodel.mats3D.mats3D_tensor import \
    map3d_tns4_to_tns2
from ibvpy.tmodel.matsXD.matsXD_cmdm.matsXD_cmdm import \
    MATSXDMicroplaneDamage
from ibvpy.tmodel.mats_eval import \
    IMATSEval
from ibvpy.mathkit.mfn.mfn_polar.mfn_polar import MFnPolar
from numpy import \
    array, ones, outer, \
    identity
from traits.api import \
    Enum, Property, cached_property, Constant, Type, \
    Int, provides

import numpy as np


@provides(IMATSEval)
class MATS2DMicroplaneDamage(MATSXDMicroplaneDamage, MATS2DEval):

    # number of spatial dimensions
    #
    n_dim = Constant(2)

    # number of components of engineering tensor representation
    #
    n_eng = Constant(3)

    # planar constraint
    stress_state = Enum("plane_strain", "plane_stress")

    # Specify the class to use for directional dependence
    mfn_class = Type(MFnPolar)

    # get the normal vectors of the microplanes
    _MPN = Property(depends_on='n_mp')

    @cached_property
    def _get__MPN(self):
        return array([[cos(alpha), sin(alpha)] for alpha in self.alpha_list])

    # get the weights of the microplanes
    _MPW = Property(depends_on='n_mp')

    @cached_property
    def _get__MPW(self):
        return ones(self.n_mp) / self.n_mp * 2

    elasticity_tensors = Property(depends_on='E, nu, stress_state')

    @cached_property
    def _get_elasticity_tensors(self):
        '''
        Intialize the fourth order elasticity tensor
        for 3D or 2D plane strain or 2D plane stress
        '''
        # ----------------------------------------------------------------------------
        # Lame constants calculated from E and nu
        # ----------------------------------------------------------------------------
        E = self.E
        nu = self.nu

        # first Lame paramter
        la = E * nu / ((1 + nu) * (1 - 2 * nu))
        # second Lame parameter (shear modulus)
        mu = E / (2 + 2 * nu)

        # -----------------------------------------------------------------------------------------------------
        # Get the fourth order elasticity and compliance tensors for the 3D-case
        # -----------------------------------------------------------------------------------------------------

        # The following lines correspond to the tensorial expression:
        # (using numpy functionality in order to avoid the loop):
        #
        # D4_e_3D = zeros((3,3,3,3),dtype=float)
        # C4_e_3D = zeros((3,3,3,3),dtype=float)
        # delta = identity(3)
        # for i in range(0,3):
        #     for j in range(0,3):
        #         for k in range(0,3):
        #             for l in range(0,3):
        #                 # elasticity tensor (cf. Jir/Baz Inelastic analysis of structures Eq.D25):
        #                 D4_e_3D[i,j,k,l] = la * delta[i,j] * delta[k,l] + \
        #                                    mu * ( delta[i,k] * delta[j,l] + delta[i,l] * delta[j,k] )
        #                 # elastic compliance tensor (cf. Simo, Computational Inelasticity, Eq.(2.7.16) AND (2.1.16)):
        #                 C4_e_3D[i,j,k,l] = (1+nu)/(2*E) * \
        #                                    ( delta[i,k] * delta[j,l] + delta[i,l]* delta[j,k] ) - \
        #                                    nu / E * delta[i,j] * delta[k,l]
        #
        # NOTE: swapaxes returns a reference not a copy!
        # (the index notation always refers to the initial indexing (i=0,j=1,k=2,l=3))
        delta = identity(3)
        delta_ijkl = outer(delta, delta).reshape(3, 3, 3, 3)
        delta_ikjl = delta_ijkl.swapaxes(1, 2)
        delta_iljk = delta_ikjl.swapaxes(2, 3)
        D4_e_3D = la * delta_ijkl + mu * (delta_ikjl + delta_iljk)
        C4_e_3D = -nu / E * delta_ijkl + \
                  (1 + nu) / (2 * E) * (delta_ikjl + delta_iljk)

        # -----------------------------------------------------------------------------------------------------
        # Get the fourth order elasticity and compliance tensors for the 2D-case
        # -----------------------------------------------------------------------------------------------------
        # 1. step: Get the (6x6)-elasticity and compliance matrices
        #          for the 3D-case:
        D2_e_3D = map3d_tns4_to_tns2(D4_e_3D)
        C2_e_3D = map3d_tns4_to_tns2(C4_e_3D)

        # 2. step: Get the (3x3)-elasticity and compliance matrices
        #          for the 2D-cases plane stress and plane strain:
        D2_e_2D_plane_stress = get_D_plane_stress(D2_e_3D)
        D2_e_2D_plane_strain = get_D_plane_strain(D2_e_3D)
        C2_e_2D_plane_stress = get_C_plane_stress(C2_e_3D)
        C2_e_2D_plane_strain = get_C_plane_strain(C2_e_3D)

        if self.stress_state == 'plane_stress':
            D2_e = D2_e_2D_plane_stress

        if self.stress_state == 'plane_strain':
            D2_e = D2_e_2D_plane_strain

        # 3. step: Get the fourth order elasticity and compliance tensors
        # for the 2D-cases plane stress and plane strain (D4.shape = (2,2,2,2))
        D4_e_2D_plane_stress = map2d_tns2_to_tns4(D2_e_2D_plane_stress)
        D4_e_2D_plane_strain = map2d_tns2_to_tns4(D2_e_2D_plane_strain)
        C4_e_2D_plane_stress = map2d_tns2_to_tns4(C2_e_2D_plane_stress)
        C4_e_2D_plane_strain = map2d_tns2_to_tns4(C2_e_2D_plane_strain)

        # -----------------------------------------------------------------------------------------------------
        # assign the fourth order elasticity and compliance tensors as return values
        # -----------------------------------------------------------------------------------------------------
        if self.stress_state == 'plane_stress':
            # print 'stress state:   plane-stress'
            D4_e = D4_e_2D_plane_stress
            C4_e = C4_e_2D_plane_stress

        if self.stress_state == 'plane_strain':
            # print 'stress state:   plane-strain'
            D4_e = D4_e_2D_plane_strain
            C4_e = C4_e_2D_plane_strain

        return D4_e, C4_e, D2_e

    def _get_explorer_config(self):
        '''Get the specific configuration of this material model in the explorer
        '''
        c = super(MATS2DMicroplaneDamage, self)._get_explorer_config()

        from ibvpy.tmodel.mats2D.mats2D_rtrace_cylinder import MATS2DRTraceCylinder

        # overload the default configuration
        c['rtrace_list'] += [
            MATS2DRTraceCylinder(name='Laterne',
                                 var_axis='time', idx_axis=0,
                                 var_surface='microplane_damage',
                                 record_on='update'),
        ]

        return c


class MATS1DMicroplaneDamage(MATS2DMicroplaneDamage):
    n_mp = Int(2)
    _MPN = Property

    @cached_property
    def _get__MPN(self):
        # microplane normals:
        return np.array([[1, 0], [0, 1]], dtype='f')

    _MPW = Property

    @cached_property
    def _get__MPW(self):
        # microplane normals:
        return np.array([1.0, 1.0], dtype='f')




from os.path import join

from ibvpy.mathkit.mfn.mfn_line.mfn_line import MFnLineArray
from ibvpy.view.reporter import RInputRecord
from scipy.optimize import newton
from traits.api import \
    Instance, Str, \
    Float, on_trait_change,\
    Interface, provides, Range, Property, Button, \
    Array, WeakRef, observe, cached_property
import bmcs_utils.api as bu
from ibvpy.view.ui import BMCSLeafNode
import sympy as sp
import numpy as np


class IDamageFn(Interface):

    def __call__(self, k):
        '''get the value of the function'''

    def diff(self, k):
        '''get the first derivative of the function'''

    def get_f_trial(self, s, k):
        '''get the map of indexes with inelastic behavior'''


class DamageFn(BMCSLeafNode):

    mats = WeakRef

    kappa_0 = bu.Float(0.0004,
                MAT=True,
                symbol="s_0",
                desc="elastic strain limit",
                unit='mm')

    E_name = Str('E')
    '''Name of the stiffness variable in the material model'''

    E = Property(bu.Float)
    def _get_E(self):
        if self.mats:
            return getattr(self.mats,self.E_name)
        else:
            return self.E_

    def get_f_trial(self, eps_eq_Em, kappa_Em):
        k_Em = np.copy(kappa_Em)
        k_Em[k_Em < self.kappa_0] = self.kappa_0
        return np.where(eps_eq_Em >= k_Em)

    plot_min = bu.Float(1e-9)
    plot_max_ = bu.Float(1e-2)
    # name of the trait controlling the plot range
    x_max_name = bu.Str('s_max')
    # if the parameter database is set get the values from there
    plot_max = Property(bu.Float)
    def _get_plot_max(self):
        if self.mats:
            return getattr(self.mats,self.x_max_name)
        else:
            return self.plot_max_

    def plot(self, ax, **kw):
        ax_omega, ax_d_omega = ax
        n_vals = 200
        kappa_range = np.linspace(self.plot_min, self.plot_max, n_vals)
        omega_range = np.zeros_like(kappa_range)
        I = kappa_range > self.kappa_0
        if len(I) > 0:
            omega_range[I] = self.__call__(kappa_range[I])
        color = kw.pop('color', 'green')
        ax_omega.plot(kappa_range, omega_range, color=color, **kw)
        ax_d_omega.plot(kappa_range, self.diff(kappa_range),
                        color='gray', linestyle='dashed', **kw)
        ax_omega.set_xlabel(r'$\kappa$ [mm]')
        ax_omega.set_ylabel(r'$\omega$ [-]')
        ax_d_omega.set_ylabel(r'$\mathrm{d} \omega / \mathrm{d} \kappa$ [-/mm]')

    def subplots(self, fig):
        ax_omega = fig.subplots(1,1)
        ax_d_omega = ax_omega.twinx()
        return ax_omega, ax_d_omega

    def update_plot(self, axes):
        self.plot(axes)

class DamageFnInjectSymbExpr(bu.InjectSymbExpr, DamageFn):
    '''
    Damage function derived symbolically
    '''
    def __call__(self, kappa):
        return self.symb.get_omega_(kappa)

    def diff(self, kappa):
        return self.symb.get_d_omega_(kappa)

    latex_eq = Str(None)

    def _repr_latex_(self):
        return self.symb.omega_._repr_latex_()

    def update_plot(self, axes):
        self.plot(axes)

class GfDamageFnSymbExpr(bu.SymbExpr):
    '''Self regularized damage function'''
    c_1, c_2 = sp.symbols('c_1, c_2', positive=True)
    E, kappa_0 = sp.symbols(r'E_b, kappa_0', positive=True)
    kappa, G_f = sp.symbols(r'kappa, G_f', positive=True)
    f = c_1 * sp.exp(-c_2*kappa)
    f_kappa_0 = sp.Piecewise(
        (1, kappa < kappa_0),
        (f.subs(kappa, kappa - kappa_0), True)
    )
    sig_s = E * kappa * f_kappa_0
    subs_c_1 = sp.solve({sp.Eq(sig_s.subs(kappa, kappa_0), E * kappa_0)}, {c_1})
    G_f_ = sp.simplify( sp.integrate(sig_s.subs(subs_c_1), (kappa,0,sp.oo)) ).factor()
    subs_c_2_1, subs_c_2_2 = sp.solve(sp.Eq(G_f, G_f_), {c_2})
    g_kappa_1 = sp.simplify( f_kappa_0.subs(c_2, subs_c_2_1).subs(c_1,1) )
    g_kappa_2 = sp.simplify( f_kappa_0.subs(c_2, subs_c_2_2).subs(c_1,1) )
    omega_1 = 1 - g_kappa_1
    omega_2 = 1 - g_kappa_2
    d_omega_1 = omega_1.diff(kappa)
    d_omega_2 = omega_2.diff(kappa)

    symb_model_params = ['kappa_0', 'E', 'G_f']

    # List of expressions for which the methods `get_`
    symb_expressions = [
        ('omega_1', ('kappa',)),
        ('d_omega_1', ('kappa',)),
        ('omega_2', ('kappa',)),
        ('d_omega_2', ('kappa',)),
    ]

@provides(IDamageFn)
class GfDamageFn(DamageFnInjectSymbExpr):
    name = 'self-regularized'
    symb_class = GfDamageFnSymbExpr

    G_f = bu.Float(0.1, MAT=True, symbol="G_\mathrm{f}", unit='N/mm',
                desc="derivative of the damage function at the onset of damage" )

    min_G_f = Property(bu.Float, depends_on='state_changed')
    def _get_min_G_f(self):
        return self.E * self.kappa_0**2 / 2

    E_ = bu.Float(10000.0, MAT=True, label="E", desc="Young's modulus")

    E = Property(bu.Float)
    def _get_E(self):
        if self.mats:
            return getattr(self.mats,self.E_name)
        else:
            return self.E_

    positive_omega = Property
    @cached_property
    def _get_positive_omega(self):
        """
        Select the correct symbolic solution for the damage function and its derivative.

        Returns a tuple of callables and the symbolic expression: (omega_func, d_omega_func, omega_expr)
        """
        omega_1 = self.symb.get_omega_1(self.kappa_0 * 1.1)
        omega_2 = self.symb.get_omega_2(self.kappa_0 * 1.1)
        if omega_1 > omega_2:
            return (self.symb.get_omega_1, self.symb.get_d_omega_1, self.symb.omega_1)
        else:
            return (self.symb.get_omega_2, self.symb.get_d_omega_2, self.symb.omega_2)

    def __call__(self, kappa):
        omega_func, _, _ = self.positive_omega
        return omega_func(kappa)

    def diff(self, kappa):
        _, d_omega_func, _ = self.positive_omega
        return d_omega_func(kappa)

    def _repr_latex_(self):
        _, _, omega_expr = self.positive_omega
        return omega_expr._repr_latex_()

    ipw_view = bu.View(
        bu.Item('kappa_0', latex=r'\kappa_0 [\mathrm{mm}]'),
        bu.Item('G_f', latex=r'G_\mathrm{f} [\mathrm{N/mm}]'),
        bu.Item('E', latex=r'E [\mathrm{MPa}]', readonly=True),
        bu.Item('min_G_f', latex=r'\min(G_\mathrm{f})', readonly=True),
        bu.Item('plot_max', readonly=True),
    )


class ExpSlopeDamageFnSymbExpr(bu.SymbExpr):

    kappa, kappa_0 = sp.symbols(r'\kappa, \kappa_0')
    kappa_f = sp.symbols(r'\kappa_\mathrm{f}')
    omega_kappa_ =  1 - (kappa_0 / kappa * sp.exp(-(kappa-kappa_0)/(kappa_f-kappa_0) ) )
    omega_ = sp.Piecewise( (0, kappa <= kappa_0), (omega_kappa_, True) )

    d_omega_ = omega_.diff(kappa)

    symb_model_params = ['kappa_0', 'kappa_f']

    # List of expressions for which the methods `get_`
    symb_expressions = [
        ('omega_', ('kappa',)),
        ('d_omega_', ('kappa',)),
    ]

@provides(IDamageFn)
class ExpSlopeDamageFn(DamageFnInjectSymbExpr):

    name = 'Exponential with slope'
    symb_class = ExpSlopeDamageFnSymbExpr

    kappa_f = bu.Float(0.001, MAT=True, symbol=r'\kappa_\mathrm{f}', unit='mm/mm',
                desc="derivative of the damage function at the onset of damage" )

    ipw_view = bu.View(
        bu.Item('kappa_0', latex=r'\kappa_0'),
        bu.Item('kappa_f', latex=r'\kappa_\mathrm{f}'),
    )


class AbaqusDamageFnSymbExpr(bu.SymbExpr):
    kappa, kappa_0 = sp.symbols(r'\kappa, \kappa_0')
    alpha, kappa_u = sp.symbols('alpha, kappa_u')
    g2_kappa_ = (kappa_0 / kappa * (1 - (1 - sp.exp(-alpha * ((kappa - kappa_0) / (kappa_u - kappa_0)))) /
                            (1 - sp.exp(-alpha))))
    g2_ = sp.Piecewise((1, kappa < kappa_0), (g2_kappa_, kappa < kappa_u), (0, True))
    omega_kappa_ =  1 - g2_
    omega_ = sp.Piecewise( (0, kappa <= kappa_0), (omega_kappa_, True) )
    d_omega_ = omega_.diff(kappa)

    symb_model_params = ['kappa_0', 'kappa_u', 'alpha']

    # List of expressions for which the methods `get_`
    symb_expressions = [
        ('omega_', ('kappa',)),
        ('d_omega_', ('kappa',)),
    ]

@provides(IDamageFn)
class AbaqusDamageFn(DamageFnInjectSymbExpr):

    name = 'Abaqus damage function'
    symb_class = AbaqusDamageFnSymbExpr

    latex_eq = Str(r'''Damage function (Abaqus)
        \begin{align}
        \omega = g(\kappa) = 1 -\left(\frac{s_0}{\kappa}\right)\left[ 1 - \frac{1 - \exp(- \alpha(\frac{\kappa - s_0}{s_u - s_0})}{1 - \exp(-\alpha)}  \right]
        \end{align}
        where $\kappa$ is the state variable representing 
        the maximum slip that occurred so far in
        in the history of loading.
        ''')

    kappa_u = bu.Float(0.003, MAT=True, symbol="kappa_u", unit='mm',
                desc="parameter of the damage function",
    )

    alpha = bu.Float(0.1, MAT=True, symbol=r"\alpha",
                  desc="parameter controlling the slope of damage",
                  unit='-')

    ipw_view = bu.View(
        bu.Item('kappa_0', latex=r'\kappa_0'),
        bu.Item('kappa_u', latex=r'\kappa_\mathrm{u}'),
        bu.Item('alpha'),
    )

class LinearDamageFnSymbExpr(bu.SymbExpr):
    kappa, kappa_0 = sp.symbols(r'\kappa, \kappa_0')
    kappa_u = sp.symbols(r'\kappa_u')
    omega_ = sp.Piecewise(
        (0, kappa < kappa_0),
        (1/(kappa_u-kappa_0)*(kappa-kappa_0), kappa < kappa_u),
        (1, True)
    )
    d_omega_ = omega_.diff(kappa)

    symb_model_params = ['kappa_0', 'kappa_u']

    # List of expressions for which the methods `get_`
    symb_expressions = [
        ('omega_', ('kappa',)),
        ('d_omega_', ('kappa',)),
    ]

@provides(IDamageFn)
class LinearDamageFn(DamageFnInjectSymbExpr):

    name = 'Linear damage function'
    symb_class = LinearDamageFnSymbExpr

    kappa_0 = 0.01

    kappa_u = bu.Float(0.03, MAT=True, symbol="kappa_u", unit='mm',
                desc="parameter of the damage function",
    )

    ipw_view = bu.View(
        bu.Item('kappa_0', latex=r'\kappa_0'),
        bu.Item('kappa_u', latex=r'\kappa_\mathrm{u}'),
    )

class WeibullDamageFnSymbExpr(bu.SymbExpr):
    kappa = sp.symbols(r'\kappa')
    m, lambda_ = sp.symbols(r'm, \lambda')
    beta_ = 1 / lambda_
    weibull_cdf = (1 - sp.exp(-(beta_ * kappa)**m))
    omega_ = sp.Piecewise(
        (0, kappa < 0),
        (weibull_cdf, True)
    )
    d_omega_ = omega_.diff(kappa)

    symb_model_params = ['m', 'lambda_']

    # List of expressions for which the methods `get_`
    symb_expressions = [
        ('omega_', ('kappa',)),
        ('d_omega_', ('kappa',)),
    ]

@provides(IDamageFn)
class WeibullDamageFn(DamageFnInjectSymbExpr):

    name = 'Linear damage function'
    symb_class = WeibullDamageFnSymbExpr

    kappa_0 = bu.Float(1e-5)
    # the inelastic regime starts right from the beginning

    lambda_ = bu.Float(0.03, MAT=True, symbol="lambda_", unit='mm',
                desc="Weibull scale parameter",
    )

    m = bu.Float(5, MAT=True, symbol="m",
                desc="Weibull shape parameter",
    )

    ipw_view = bu.View(
        bu.Item('lambda_', latex=r'\lambda'),
        bu.Item('m', latex=r'm'),
    )

@provides(IDamageFn)
class LiDamageFn(DamageFn):

    name = 'Two parameter damage'

    latex_eq = Str(r'''Damage function (Li)
        \begin{align}
        \omega = g(\kappa) = \frac{\alpha_1}{1 + \exp(-\alpha_2 \kappa + 6 )}
        \end{align}
        where $\kappa$ is the state variable representing 
        the maximum slip that occurred so far in
        in the history of loading.
        ''')

    alpha_1 = bu.Float(value=1, MAT=True,
                    symbol=r'\alpha_1',
                    unit='-',
                    desc="parameter controlling the shape of the damage function")

    alpha_2 = bu.Float(2000., MAT=True,
                    symbol=r'\alpha_2',
                    unit='-',
                    desc="parameter controlling the shape of the damage function")

    def __call__(self, kappa):
        alpha_1 = self.alpha_1
        alpha_2 = self.alpha_2
        s_0 = self.s_0
        omega = np.zeros_like(kappa, dtype=np.float64)
        d_idx = np.where(kappa >= s_0)[0]
        k = kappa[d_idx]
        omega[d_idx] = 1. / \
            (1. + np.exp(-1. * alpha_2 * (k - s_0) + 6.)) * alpha_1
        return omega

    def diff(self, kappa):
        alpha_1 = self.alpha_1
        alpha_2 = self.alpha_2
        s_0 = self.s_0
        return ((alpha_1 * alpha_2 *
                 np.exp(-1. * alpha_2 * (kappa - s_0) + 6.)) /
                (1 + np.exp(-1. * alpha_2 * (kappa - s_0) + 6.)) ** 2)

    ipw_view = bu.View(
        bu.Item('s_0'),
        bu.Item('alpha_1', editor=bu.FloatRangeEditor(low=0,high=1)),
        bu.Item('alpha_2'),
    )


@provides(IDamageFn)
class MultilinearDamageFn(DamageFn):

    name = 'Multilinear damage function'

    s_data = bu.Str('1', tooltip='Comma-separated list of strain values',
                 MAT=True, unit='mm', symbol='s',
                 desc='slip values',
                 )

    omega_data = bu.Str('1', tooltip='Comma-separated list of damage values',
                     MAT=True, unit='-', symbol=r'\omega',
                     desc='shear stress values',
                     )

    s_omega_table = Property

    def _set_s_omega_table(self, data):
        s_data, omega_data = data
        if len(s_data) != len(omega_data):
            raise ValueError('s array and tau array must have the same size')
        self.damage_law.set(xdata=s_data,
                            ydata=omega_data)

    @observe('state_changed')
    def _update_damage_law(self, event=None):
        s_data = np.fromstring(self.s_data, dtype=np.float64, sep=',')
        omega_data = np.fromstring(self.omega_data, dtype=np.float64, sep=',')
        s_data = np.hstack([[0, self.s_0], s_data])
        omega_data = np.hstack([[0, 0], omega_data])
        n = np.min([len(s_data), len(omega_data)])
        self.damage_law.set(xdata=s_data[:n], ydata=omega_data[:n])

    damage_law = Instance(MFnLineArray)

    def _damage_law_default(self):
        return MFnLineArray(
            xdata=[0.0, 1.0],
            ydata=[0.0, 0.0],
            plot_diff=False)

    def __call__(self, kappa):
        shape = kappa.shape
        return self.damage_law(kappa.flatten()).reshape(*shape)

    def diff(self, kappa):
        shape = kappa.shape
        return self.damage_law.diff(kappa.flatten()).reshape(*shape)

    ipw_view = bu.View(
        bu.Item('s_0'),
        bu.Item('s_data'),
        bu.Item('omega_data'),
    )

class GfDamageFn2(DamageFn):
    '''Class defining the damage function coupled with the fracture
    energy of a cohesive crack model.
    '''
    name = 'damage function Gf'

    L_s = bu.Float(1.0, MAT=True, label="L_s",
                   desc="Length of the softening zone")

    E_ = bu.Float(34000.0, MAT=True, label="E", desc="Young's modulus")

    f_t = bu.Float(4.5, MAT=True, label="f_t", desc="Tensile strength")

    f_t_Em = Array(np.float64, value=None)

    G_f = bu.Float(0.004, MAT=True, label="G_f", desc="Fracture energy",)

    s_0 = Property(bu.Float)
    def _get_s_0(self):
        return self.f_t / self.E

    eps_ch = Property(bu.Float)

    def _get_eps_ch(self):
        return self.G_f / self.f_t

    ipw_view = bu.View(
                bu.Item('L_s', latex=r'L_\mathrm{s}'),
                bu.Item('f_t', latex=r'f_\mathrm{t}'),
                bu.Item('G_f', latex=r'G_\mathrm{f}'),
                bu.Item('E', readonly=True),
                bu.Item('s_0', latex=r's_0', readonly=True),
            )

    def __call__(self, kappa):
        L_s = self.L_s
        f_t = self.f_t
        G_f = self.G_f
        E = self.E
        s_0 = self.s_0
        return (
            1 - f_t * np.exp(-f_t * (kappa - s_0) * L_s / G_f)
            / (E * kappa)
        )

    def diff(self, kappa):
        L_s = self.L_s
        f_t = self.f_t
        G_f = self.G_f
        E = self.E
        s_0 = self.s_0
        return (
            f_t * np.exp(L_s * (s_0 - kappa) * f_t / G_f)
            / (E * G_f * kappa**2) * (G_f + L_s * kappa * f_t)
        )

@provides(IDamageFn)
class FRPDamageFn(DamageFn):

    name = 'FRP damage function'

    B = bu.Float(10.4,
              MAT=True,
              symbol="B",
              unit='mm$^{-1}$',
              desc="parameter controlling the damage maximum stress level")

    Gf = bu.Float(1.19,
               MAT=True,
               symbol="G_\mathrm{f}",
               unit='N/mm',
               desc="fracture energy")

    E_bond = bu.Float(0.0)

    E_b = Property(Float)

    def _get_E_b(self):
        return self.mats.E_b

    def _set_E_b(self, value):
        self.E_bond = value
        self.mats.E_b = value

    @observe('B, Gf')
    def _update_dependent_params(self, event=None):
        self.E_b = 1.734 * self.Gf * self.B ** 2.0
        # calculation of s_0, implicit function solved using Newton method

        def f_s(s_0): return s_0 / \
            (np.exp(- self.B * s_0) - np.exp(-2.0 * self.B * s_0)) - \
            2.0 * self.B * self.Gf / self.E_b
        self.s_0 = newton(f_s, 0.00000001, tol=1e-5, maxiter=20)

    def __call__(self, kappa):

        b = self.B
        Gf = self.Gf
        Eb = self.E_b  # 1.734 * Gf * b**2
        s_0 = self.s_0
        # calculation of s_0, implicit function solved using Newton method

#         def f_s(s_0): return s_0 / \
#             (np.exp(-b * s_0) - np.exp(-2.0 * b * s_0)) - 2.0 * b * Gf / Eb
#         s_0 = newton(f_s, 0.00000001, tol=1e-5, maxiter=20)

        omega = np.zeros_like(kappa, dtype=np.float64)
        I = np.where(kappa >= s_0)[0]
        kappa_I = kappa[I]

        omega[I] = 1 - \
            (2.0 * b * Gf * (np.exp(-b * kappa_I)
                             - np.exp(-2.0 * b * kappa_I))) / (kappa_I * Eb)

        return omega

    def diff(self, kappa):

        nz_ix = np.where(kappa != 0.0)[0]

        b = self.B
        Gf = self.Gf
        Eb = 1.734 * Gf * b**2

        domega_dkappa = np.zeros_like(kappa)
        kappa_nz = kappa[nz_ix]
        domega_dkappa[nz_ix] = (
            (2.0 * b * Gf *
             (np.exp(-b * kappa_nz) -
              np.exp(-2.0 * b * kappa_nz))
             ) / (Eb * kappa_nz**2.0) -
            (2.0 * b * Gf *
             (-b * np.exp(-b * kappa_nz) +
              2.0 * b * np.exp(-2.0 * b * kappa_nz))) /
            (Eb * kappa_nz)
        )
        return domega_dkappa

    latex_eq = r'''Damage function (FRP)
        \begin{align}
        \omega = g(\kappa) = 
        1 - {\frac {{\exp(-2\,Bs)}-{\exp(-Bs)}}{Bs}}
        \end{align}
        where $\kappa$ is the state variable representing 
        the maximum slip that occurred so far in
        in the history of loading.
        '''

    ipw_view = bu.View(
        bu.Item('s_0', readonly=True),
        bu.Item('E_bond', readonly=True),
        bu.Item('B'),
        bu.Item('Gf'),
    )




r'''

Simulator implementation
========================

'''
import bmcs_utils.api as bu

from traits.api import \
    Instance, on_trait_change, Str, \
    Property, cached_property, Bool, List, provides
from ibvpy.view import BMCSTreeNode, itags_str, IBMCSModel

import traits.api as tr

from .i_tloop import ITLoop
from .i_tstep import ITStep
from .tline_mixin import TLineMixIn

@provides(IBMCSModel)
class Simulator(BMCSTreeNode, TLineMixIn):
    r'''Base class for simulators included in the BMCS Tool Suite.
    It implements the state dependencies within the simulation tree.
    It handles also the communication between the simulation and
    the user interface in several modes of interaction.
    '''
    name = 'simulator'

    @tr.observe('state_changed')
    def _model_structure_changed(self, event=None):
        self.tloop.restart = True

    #=========================================================================
    # TIME LOOP
    #=========================================================================

    tloop = Property(Instance(ITLoop), depends_on='state_changed')
    r'''Time loop constructed based on the current model.
    '''
    @cached_property
    def _get_tloop(self):
        return self.tstep.tloop_type(tstep=self.tstep,
                                     tline=self.tline)

    def __init__(self, tstep, *args, **kw):
        self.tstep = tstep
        super(Simulator, self).__init__(*args, **kw)

    tstep = tr.WeakRef(ITStep)

    hist = tr.Property

    def _get_hist(self):
        return self.tstep.hist

    def run(self):
        r'''Run a thread if it does not exist - do nothing otherwise
        '''
        self.tloop()
        return

    interrupt = tr.DelegatesTo('tloop')

    def reset(self):
        self.tloop.reset()

    ipw_view = bu.View(
        run_method='tloop',
        reset_method='reset',
        interrupt_var='interrupt',
        time_var = 't',
        time_max='t_max',
    )



from ibvpy.fets.fets import FETSEval
from numpy import \
    array, zeros, int_, float64, ix_, dot, linspace, hstack, vstack, arange, \
    identity
from scipy.linalg import \
    inv
from traits.api import \
    Array, Bool, Callable, Enum, Float, HasTraits, Interface, \
    Instance, Int, Trait, Str, Enum, Callable, List, TraitDict, Any, \
    on_trait_change, Tuple, WeakRef, Delegate, Property, cached_property


#-------------------------------------------------------------------------
# FETS2D4T - 4 nodes iso-parametric quadrilateral element (2D, linear, Lagrange family)
#-------------------------------------------------------------------------
class FETS2D4Q4T(FETSEval):

    debug_on = True

    # Dimensional mapping
    dim_slice = slice(0, 2)

    # Order of node positions for the formulation of shape function
    #
    dof_r = Array(value=[[-1, -1], [1, -1], [1, 1], [-1, 1]])
    geo_r = Array(value=[[-1, -1], [1, -1], [1, 1], [-1, 1]])

    n_e_dofs = Int(4)
    t = Float(1.0, label='thickness')

    # Integration parameters
    #
    ngp_r = Int(2)
    ngp_s = Int(2)

    # Corner nodes are used for visualization
    vtk_r = Array(value=[[-1., -1.], [1., -1.], [1., 1.], [-1., 1.]])
    vtk_cells = [[0, 1, 2, 3]]
    vtk_cell_types = 'Quad'

    # vtk_point_ip_map = [0,1,3,2]
    n_nodal_dofs = Int(1)

    #---------------------------------------------------------------------
    # Method required to represent the element geometry
    #---------------------------------------------------------------------
    def get_N_geo_mtx(self, r_pnt):
        '''
        Return the value of shape functions for the specified local coordinate r
        '''
        cx = array(self.geo_r, dtype='float64')
        Nr = array([[1 / 4. * (1 + r_pnt[0] * cx[i, 0]) * (1 + r_pnt[1] * cx[i, 1])
                     for i in range(0, 4)]])
        return Nr

    def get_dNr_geo_mtx(self, r_pnt):
        '''
        Return the matrix of shape function derivatives.
        Used for the conrcution of the Jacobi matrix.

        @TODO - the B matrix is used
        just for uniaxial bar here with a trivial differential
        operator.
        '''
        cx = array(self.geo_r, dtype='float64')
        dNr_geo = array([[1 / 4. * cx[i, 0] * (1 + r_pnt[1] * cx[i, 1]) for i in range(0, 4)],
                         [1 / 4. * cx[i, 1] * (1 + r_pnt[0] * cx[i, 0]) for i in range(0, 4)]])
        return dNr_geo

    #---------------------------------------------------------------------
    # Method delivering the shape functions for the field variables and their derivatives
    #---------------------------------------------------------------------
    def get_N_mtx(self, r_pnt):
        '''
        Returns the matrix of the shape functions used for the field approximation
        containing zero entries. The number of rows corresponds to the number of nodal
        dofs. The matrix is evaluated for the specified local coordinate r.
        '''
        Nr_geo = self.get_N_geo_mtx(r_pnt)
        I_mtx = identity(self.n_nodal_dofs, float)
        N_mtx_list = [I_mtx * Nr_geo[0, i] for i in range(0, Nr_geo.shape[1])]
        N_mtx = hstack(N_mtx_list)
        return N_mtx

    def get_dNr_mtx(self, r_pnt):
        '''
        Return the derivatives of the shape functions
        '''
        return self.get_dNr_geo_mtx(r_pnt)

    def get_B_mtx(self, r_pnt, X_mtx):
        J_mtx = self.get_J_mtx(r_pnt, X_mtx)
        dNr_mtx = self.get_dNr_mtx(r_pnt)
        dNx_mtx = dot(inv(J_mtx), dNr_mtx)
        Bx_mtx = zeros((2, 4), dtype='float64')
        for i in range(0, 4):
            Bx_mtx[0, i] = dNx_mtx[0, i]
            Bx_mtx[1, i] = dNx_mtx[1, i]
        return Bx_mtx

#----------------------- example --------------------


def run_example():
    from ibvpy.api import \
        TStepper as TS, RTDofGraph, RTraceDomainListField, \
        RTraceDomainListInteg, TLoop, \
        TLine, BCDof, IBVPSolve as IS, DOTSEval
    from ibvpy.tmodel.mats2D.mats2D_conduction.mats2D_conduction import MATS2DConduction

    from ibvpy.api import BCDofGroup
    fets_eval = FETS2D4Q4T(mats_eval=MATS2DConduction(k=1.))

    print(fets_eval.vtk_node_cell_data)

    from ibvpy.mesh.fe_grid import FEGrid
    from ibvpy.mesh.fe_refinement_grid import FERefinementGrid
    from ibvpy.mesh.fe_domain import FEDomain
    from mathkit.mfn import MFnLineArray

    # Discretization
    fe_grid = FEGrid(coord_max=(1., 1., 0.),
                     shape=(2, 2),
                     fets_eval=fets_eval)

    tstepper = TS(sdomain=fe_grid,
                  bcond_list=[BCDofGroup(var='u', value=0., dims=[0],
                                         get_dof_method=fe_grid.get_left_dofs),
                              #                                   BCDofGroup( var='u', value = 0., dims = [1],
                              # get_dof_method = fe_grid.get_bottom_dofs ),
                              BCDofGroup(var='u', value=.005, dims=[0],
                                             get_dof_method=fe_grid.get_top_right_dofs)],
                  rtrace_list=[
                      #                     RTraceDomainListField(name = 'Damage' ,
                      #                                    var = 'omega', idx = 0,
                      #                                    record_on = 'update',
                      #                                    warp = True),
                      #                     RTraceDomainListField(name = 'Displacement' ,
                      #                                    var = 'u', idx = 0,
                      #                                    record_on = 'update',
                      #                                    warp = True),
                      #                    RTraceDomainListField(name = 'N0' ,
                      #                                      var = 'N_mtx', idx = 0,
                      # record_on = 'update')
                  ]
                  )

    # Add the time-loop control
    tloop = TLoop(tstepper=tstepper, debug=False,
                  tline=TLine(min=0.0, step=1.0, max=1.0))

    tloop.eval()
    # Put the whole thing into the simulation-framework to map the
    # individual pieces of definition into the user interface.
    #
    # from ibvpy.plugins.ibvpy_app import IBVPyApp
    # app = IBVPyApp( ibv_resource = tloop )
    # app.main()


if __name__ == '__main__':
    run_example()



'''
Created on Feb 14, 2019

@author: rch
'''

from ibvpy.tmodel import MATSEval
import numpy as np
from ibvpy.sim.i_tmodel import ITModel
import traits.api as tr
import bmcs_utils.api as bu

@tr.provides(ITModel)
class MATS1D5Elastic(MATSEval):

    node_name = "multilinear bond law"

    E_s = bu.Float(100.0, tooltip='Shear stiffness of the interface [MPa]',
                   MAT=True, unit='MPa', symbol='E_\mathrm{s}',
                   desc='Shear-modulus of the interface',
                   auto_set=True, enter_set=True)

    E_n = bu.Float(100.0, tooltip='Normal stiffness of the interface [MPa]',
                   MAT=True, unit='MPa', symbol='E_\mathrm{n}',
                   desc='Normal stiffness of the interface',
                   auto_set=False, enter_set=True)

    state_var_shapes = {}

    D_rs = tr.Property(depends_on='E_n,E_s')

    @tr.cached_property
    def _get_D_rs(self):
        return np.array([[self.E_s, 0],
                         [0, self.E_n]], dtype=np.float64)

    def get_corr_pred(self, u_r, tn1):
        tau = np.einsum(
            'rs,...s->...r',
            self.D_rs, u_r)
        grid_shape = tuple([1 for _ in range(len(u_r.shape[:-1]))])
        D = self.D_rs.reshape(grid_shape + (2, 2))
        return tau, D




import copy

from ibvpy.tmodel.mats_eval import \
    MATSEval
from traits.api import \
    Constant, Property, cached_property, Array,\
    Dict, Str, Callable

from bmcs_utils.api import Float, Int
import numpy as np


class MATSXDEval(MATSEval):
    '''Base class for elastic models.
    '''

    n_dims = Constant(Float)
    '''Number of spatial dimensions of an integration 
    cell for the material model
    '''

    E = Float(34e+3,
              label="E",
              desc="Young's Modulus",
              MAT=True)

    nu = Float(0.2,
               label='nu',
               desc="Poison's ratio",
               MAT=True)

    eps_max = Float(0.03, ALG=True)

    n_eps = Int(100, ALG=True)

    def subplots(self, fig):
        ax_sig = fig.subplots(1,1)
        ax_d_sig = ax_sig.twinx()
        return ax_sig, ax_d_sig

    def update_plot(self, axes):
        ax_sig, ax_d_sig = axes
        eps_max = self.eps_max
        n_eps = 100
        eps11_range = np.linspace(1e-9,eps_max,n_eps)
        eps_range = np.zeros((n_eps, self.n_dims, self.n_dims))
        eps_range[:,0,0] = eps11_range
        state_vars = { var : np.zeros( (n_eps,) + shape )
            for var, shape in self.state_var_shapes.items()
        }
        sig_range, D_range = self.get_corr_pred(eps_range, 1, **state_vars)
        sig11_range = sig_range[:,0,0]
        ax_sig.plot(eps11_range, sig11_range,color='blue')
        d_sig1111_range = D_range[...,0,0,0,0]
        ax_d_sig.plot(eps11_range, d_sig1111_range, linestyle='dashed', color='gray')
        ax_sig.set_xlabel(r'$\varepsilon_{11}$ [-]')
        ax_sig.set_ylabel(r'$\sigma_{11}$ [MPa]')
        ax_d_sig.set_ylabel(r'$\mathrm{d} \sigma_{11} / \mathrm{d} \varepsilon_{11}$ [MPa]')
        ax_d_sig.plot(eps11_range[:-1],
                    (sig11_range[:-1]-sig11_range[1:])/(eps11_range[:-1]-eps11_range[1:]),
                    color='orange', linestyle='dashed')


    def _get_lame_params(self):
        # First Lame parameter (bulk modulus)
        la = self.E * self.nu / ((1. + self.nu) * (1. - 2. * self.nu))
        # second Lame parameter (shear modulus)
        mu = self.E / (2. + 2. * self.nu)
        return la, mu

    D_abef = Property(Array, depends_on='E, nu')
    '''Material stiffness - rank 4 tensor
    '''
    @cached_property
    def _get_D_abef(self):
        la, mu = self._get_lame_params()
        delta = np.identity(self.n_dims)
        return (
            np.einsum(',ij,kl->ijkl', la, delta, delta) +
            np.einsum(',ik,jl->ijkl', mu, delta, delta) +
            np.einsum(',il,jk->ijkl', mu, delta, delta)
        )

    def get_corr_pred(self, eps_ab, tn1, **Eps):
        '''
        Corrector predictor computation.
        @param eps_Emab input variable - strain tensor
        '''
        sigma_ab = np.einsum(
            'abcd,...cd->...ab', self.D_abef, eps_ab
        )
        Em_len = len(eps_ab.shape) - 2
        new_shape = tuple([1 for _ in range(Em_len)]) + self.D_abef.shape
        D_abef = self.D_abef.reshape(*new_shape)
        return sigma_ab, D_abef

    #=========================================================================
    # Response variables
    #=========================================================================
    def get_eps_ab(self, eps_ab, tn1, **Eps):
        return eps_ab

    def get_sig_ab(self, eps_ab, tn1, **Eps):
        '''
        Get the stress tensor
        :param eps_ab: strain tensor with ab indexes
        :param tn1: time - can be used for time dependent properties
        :param Eps: dictionary of state variables corresponding to the
         specification in state_var_shapes
        :return: stress vector with with the shape of input and state variables + ab
         spatial indexes.
        '''
        return self.get_sig(eps_ab, tn1, **Eps)

    def get_state_var(self, var_name, eps_ab, tn1, **Eps):
        '''
        Acess to state variables used for visualization
        :param var_name: string with the name of the state variable
        :param eps_ab: strain tensor (not used)
        :param tn1: time variable (not used)
        :param Eps: dictionary of state variables corresponding to the
         specification in state_var_shapes
        :return: Array with the state variable with the shape corresponding
         to the level of the problem
        '''
        return Eps[var_name]

    var_dict = Property(Dict(Str, Callable))
    '''Dictionary of functions that deliver the response/state variables 
    of a material model
    '''
    @cached_property
    def _get_var_dict(self):
        var_dict = super(MATSXDEval, self)._get_var_dict()
        var_dict.update(eps_ab=self.get_eps_ab,
                        sig_ab=self.get_sig_ab)
        var_dict.update({
            var_name : lambda eps_ab, tn1, **Eps : self.get_state_var(
                var_name, eps_ab, tn1, **Eps
            )
            for var_name in self.state_var_shapes.keys()
        })
        return var_dict



'''
Created on Feb 14, 2019

@author: rch
'''

from ibvpy.tmodel import MATSEval
from ibvpy.sim.i_tmodel import ITModel

import numpy as np
import traits.api as tr
import bmcs_utils.api as bu


@tr.provides(ITModel)
class MATS1D5DPCumPress(MATSEval):

    node_name = 'Pressure sensitive cumulative damage plasticity'

    E_N = bu.Float(30000, label='E_N',
                   desc='Normal stiffness of the interface',
                   MAT=True,
                   enter_set=True, auto_set=False)

    E_T = bu.Float(12900, label='E_T',
                   desc='Shear modulus of the interface',
                   MAT=True,
                   enter_set=True, auto_set=False)

    gamma = bu.Float(55.0, label='gamma',
                     desc='Kinematic Hardening Modulus',
                     MAT=True,
                     enter_set=True, auto_set=False)

    K = bu.Float(11, label='K',
                 desc='Isotropic hardening modulus',
                 MAT=True,
                 enter_set=True, auto_set=False)

    S = bu.Float(0.005, label='S',
                 desc='Damage accumulation parameter',
                 MAT=True,
                 enter_set=True, auto_set=False)

    r = bu.Float(1, label='r',
                 desc='Damage accumulation parameter',
                 MAT=True,
                 enter_set=True, auto_set=False)

    c = bu.Float(1, Label='c',
                 desc='Damage accumulation parameter',
                 MAT=True,
                 enter_set=True, auto_set=False)
    m = bu.Float(0.3, label='m',
                 desc='Lateral Pressure Coefficient',
                 MAT=True,
                 enter_set=True, auto_set=False)

    tau_bar = bu.Float(4.2, label='tau_bar',
                       desc='Reversibility limit',
                       MAT=True,
                       enter_set=True, auto_set=False)

    state_var_shapes = dict(s_pi=(),
                            alpha=(),
                            z=(),
                            omega=())

    D_rs = tr.Property(depends_on='E_N,E_T')

    @tr.cached_property
    def _get_D_rs(self):
        print('recalculating D_rs')
        return np.array([[self.E_T, 0],
                         [0, self.E_N]], dtype=np.float64)

    def init(self, s_pi, alpha, z, omega):
        r'''
        Initialize the state variables.
        '''
        s_pi[...] = 0
        alpha[...] = 0
        z[...] = 0
        omega[...] = 0

    algorithmic = tr.Bool(True)

    def get_corr_pred(self, u_r, t_n, s_pi, alpha, z, omega):

        s = u_r[..., 0]
        w = u_r[..., 1]
        # For normal direction
        H_w_N = np.array(w <= 0.0, dtype=np.float64)
        E_alg_N = H_w_N * self.E_N
        sig_N = E_alg_N * w
        # For tangential
        # Y = 0.5 * self.E_T * (u_T - s_pi)**2
        tau_pi_trial = self.E_T * (s - s_pi)
        Z = self.K * z
        X = self.gamma * alpha
        f = np.fabs(tau_pi_trial - X) - Z - self.tau_bar + self.m * sig_N
        # Identify inelastic material points
        # @todo: consider the usage of np.where()
        I = f > 1e-6
        sig_T = (1 - omega) * self.E_T * (s - s_pi)
        # Return mapping
        delta_lambda_I = (
            f[I] / (self.E_T / (1 - omega[I]) + self.gamma + self.K)
        )
        # Update all state variables
        s_pi[I] += (delta_lambda_I * 
                    np.sign(tau_pi_trial[I] - X[I]) / (1 - omega[I]))
        Y = 0.5 * self.E_T * (s - s_pi) ** 2
        omega[I] += (
            delta_lambda_I * 
            (1 - omega[I]) ** self.c * (Y[I] / self.S) ** self.r * (self.tau_bar / (self.tau_bar - self.m * sig_N[I]))
        )
        sig_T[I] = (1 - omega[I]) * self.E_T * (s[I] - s_pi[I])
        alpha[I] += delta_lambda_I * np.sign(tau_pi_trial[I] - X[I])
        z[I] += delta_lambda_I
        # Unloading stiffness
     
        E_alg_T = (1 - omega) * self.E_T
       
        # Consistent tangent operator
        if False:
            E_alg_T = (
                (1 - omega) * self.E_T - 
                (1 - omega) * self.E_T ** 2 / 
                (self.E_T + (self.gamma + self.K) * (1 - omega)) - 
                ((1 - omega) ** self.c * (self.E_T ** 2) * ((Y / self.S) ** self.r)
                 * np.sign(tau_pi_trial - X) * (s - s_pi)) / 
                ((self.E_T / (1 - omega)) + self.gamma + self.K)
            )

        
        if False:
            #print('DONT COME HERE')
            E_alg_T = (
                (1 - omega) * self.E_T - 
                ((self.E_T ** 2 * (1 - omega)) / 
                 (self.E_T + (self.gamma + self.K) * (1 - omega)))
                -
                ((1 - omega) ** self.c * 
                 (Y / self.S) ** self.r * 
                    self.E_T ** 2 * (s - s_pi) * self.tau_bar / 
                    (self.tau_bar - self.m * sig_N) * np.sign(tau_pi_trial - X)) / 
                (self.E_T / (1 - omega) + self.gamma + self.K)
            )

        sig = np.zeros_like(u_r)
        sig[..., 0] = sig_T
        sig[..., 1] = sig_N
        E_TN = np.einsum('abEm->Emab',
                         np.array(
                             [
                                 [E_alg_T, np.zeros_like(E_alg_T)],
                                 [np.zeros_like(E_alg_N), E_alg_N]
                             ])
                         )
#         print(' sig_N =',  sig_N)
        abc = open('sigNfortau10000lp5000.txt', 'a+', newline='\n')
        for e in range(len(sig_N)):
            abc.write('%f ' % sig_N[e][0])
        abc.write('\n')
        abc.close()
        # print('s_pi=', s_pi)
        return sig, E_TN

    def _get_var_dict(self):
        var_dict = super(MATS1D5DPCumPress, self)._get_var_dict()
        var_dict.update(
            slip=self.get_slip,
            s_el=self.get_s_el,
            shear=self.get_shear,
            omega=self.get_omega,
            s_pi=self.get_s_pi,
            alpha=self.get_alpha,
            z=self.get_z
        )
        return var_dict

    def get_slip(self, u_r, tn1, **state):
        return self.get_eps(u_r, tn1)[..., 0]

    def get_shear(self, u_r, tn1, **state):
        return self.get_sig(u_r, tn1, **state)[..., 0]

    def get_omega(self, u_r, tn1, s_pi, alpha, z, omega):
        return omega

    def get_s_pi(self, u_r, tn1, s_pi, alpha, z, omega):
        return s_pi

    def get_alpha(self, u_r, tn1, s_pi, alpha, z, omega):
        return alpha

    def get_z(self, u_r, tn1, s_pi, alpha, z, omega):
        return z

    def get_s_el(self, u_r, tn1, **state):
        s = self.get_slip(u_r, tn1, **state)
        s_p = self.get_s_pi(u_r, tn1, **state)
        s_e = s - s_p
        return s_e

#     def get_sig_N(self, u_r, tn1, **state):
#         return self.get_sig(u_r, tn1, **state)[..., 1]

    tree_view = bu.View(
        bu.Item('E_N'),
        bu.Item('E_T'),
        bu.Item('gamma'),
        bu.Item('K'),
        bu.Item('S'),
        bu.Item('r'),
        bu.Item('c'),
        bu.Item('m'),
        bu.Item('tau_bar'),
        bu.Item('D_rs', style='readonly')
    )

    traits_view = tree_view




'''
Created on 29.03.2017

@author: abaktheer

Microplane damage model 2D - Jirasek [1999]
'''

from numpy import \
    array, einsum, identity, sqrt
from traits.api import \
    Constant, provides, \
    Property, cached_property

from bmcs_utils.api import Float, View, Item

from ibvpy.tmodel.mats2D.mats2D_eval import MATS2DEval
from ibvpy.tmodel.matsXD.vmatsXD_eval import MATSXDEval
import numpy as np
from ibvpy.sim.i_tmodel import ITModel
import traits.api as tr
from ibvpy.tmodel.mats2D.mats2D_microplane.vmats2D_calibration_mixin_Gf import \
    MATS2DCalibrationMixinGf

@provides(ITModel)
class MATS2DMplDamageEEQ(MATS2DEval, MATS2DCalibrationMixinGf):

    epsilon_0 = Float(59e-6,
                      label="a",
                      desc="Lateral pressure coefficient",
                      MAT=True)

    epsilon_f = Float(250e-6,
                      label="a",
                      desc="Lateral pressure coefficient",
                      MAT=True)

    c_T = Float(0.00,
                label="a",
                desc="Lateral pressure coefficient",
                MAT=True)

    ipw_view = View(
        *MATS2DEval.ipw_view.content,
        Item('epsilon_0'),
        Item('epsilon_f'),
        Item('c_T'),
    )

    state_var_shapes = tr.Property(tr.Dict(), depends_on='n_mp')
    '''Dictionary of state variable entries with their array shapes.
    '''

    @cached_property
    def _get_state_var_shapes(self):
        return dict(kappa=(self.n_mp,),
                    omega=(self.n_mp,))

    #-------------------------------------------------------------------------
    # MICROPLANE-Kinematic constraints
    #-------------------------------------------------------------------------

    # get the dyadic product of the microplane normals
    _MPNN = Property(depends_on='n_mp')

    @cached_property
    def _get__MPNN(self):
        # dyadic product of the microplane normals

        MPNN_nij = einsum('ni,nj->nij', self._MPN, self._MPN)
        return MPNN_nij

    # get the third order tangential tensor (operator) for each microplane
    _MPTT = Property(depends_on='n_mp')

    @cached_property
    def _get__MPTT(self):
        # Third order tangential tensor for each microplane
        delta = identity(2)
        MPTT_nijr = 0.5 * (einsum('ni,jr -> nijr', self._MPN, delta) +
                           einsum('nj,ir -> njir', self._MPN, delta) - 2.0 *
                           einsum('ni,nj,nr -> nijr', self._MPN, self._MPN, self._MPN))
        return MPTT_nijr

    def _get_e_na(self, eps_ab):
        r'''
        Projection of apparent strain onto the individual microplanes
        '''
        e_ni = einsum(
            'nb,...ba->...na',
            self._MPN, eps_ab
        )
        return e_ni

    def _get_e_N_n(self, e_na):
        r'''
        Get the normal strain array for each microplane
        '''
        e_N_n = einsum(
            '...na, na->...n',
            e_na, self._MPN
        )
        return e_N_n

    def _get_e_equiv_n(self, e_na):
        r'''
        Returns a list of the microplane equivalent strains
        based on the list of microplane strain vectors
        '''
        # magnitude of the normal strain vector for each microplane
        e_N_n = self._get_e_N_n(e_na)
        # positive part of the normal strain magnitude for each microplane
        e_N_pos_n = (np.abs(e_N_n) + e_N_n) / 2.0
        # normal strain vector for each microplane
        e_N_na = einsum('...n,ni -> ...ni', e_N_n, self._MPN)
        # tangent strain ratio
        c_T = self.c_T
        # tangential strain vector for each microplane
        e_T_na = e_na - e_N_na
        # squared tangential strain vector for each microplane
        e_TT_n = einsum('...ni,...ni -> ...n', e_T_na, e_T_na)
        # equivalent strain for each microplane
        e_equiv_n = sqrt(e_N_pos_n * e_N_pos_n + c_T * e_TT_n)
        return e_equiv_n

    def update_state_variables(self, eps_ab, kappa_n, omega_n):
        e_na = self._get_e_na(eps_ab)
        eps_eq_n = self._get_e_equiv_n(e_na)
        f_trial_n = eps_eq_n - self.epsilon_0
        I = np.where(f_trial_n > 0)
        k_n = np.max(np.array([kappa_n[I], eps_eq_n[I]]), axis=0)
        kappa_n[I] = k_n
        omega_n[I] = self._get_omega(k_n)

    def _get_omega(self, kappa_Emn):
        '''
        Return new value of damage parameter
        @par bbam kappa:
        '''
        omega_Emn = np.zeros_like(kappa_Emn)
        epsilon_0 = self.epsilon_0
        epsilon_f = self.epsilon_f
        kappa_idx = np.where(kappa_Emn > epsilon_0)
        omega_Emn[kappa_idx] = (
            1.0 - (epsilon_0 / kappa_Emn[kappa_idx] *
                   np.exp(-1.0 * (kappa_Emn[kappa_idx] - epsilon_0) /
                          (epsilon_f - epsilon_0))
                   ))
        return omega_Emn

    def _get_phi_Emab(self, kappa_Emn):
        # Returns the 2nd order damage tensor 'phi_mtx'
        # scalar integrity factor for each microplane
        phi_Emn = np.sqrt(1.0 - self._get_omega(kappa_Emn))
        # integration terms for each microplanes
        phi_Emab = einsum('...n,n,nab->...ab', phi_Emn, self._MPW, self._MPNN)
        return phi_Emab

    def _get_beta_Emabcd(self, phi_Emab):
        '''
        Returns the 4th order damage tensor 'beta4' using sum-type symmetrization
        (cf. [Jir99], Eq.(21))
        '''
        delta = identity(2)
        beta_Emijkl = 0.25 * (einsum('...ik,jl->...ijkl', phi_Emab, delta) +
                              einsum('...il,jk->...ijkl', phi_Emab, delta) +
                              einsum('...jk,il->...ijkl', phi_Emab, delta) +
                              einsum('...jl,ik->...ijkl', phi_Emab, delta))

        return beta_Emijkl

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------

    def get_corr_pred(self, eps_ab, tn1, kappa, omega):

        self.update_state_variables(eps_ab, kappa, omega)

        #----------------------------------------------------------------------
        # if the regularization using the crack-band concept is on calculate the
        # effective element length in the direction of principle strains
        #----------------------------------------------------------------------
        # if self.regularization:
        #    h = self.get_regularizing_length(sctx, eps_app_eng)
        #    self.phi_fn.h = h

        #------------------------------------------------------------------
        # Damage tensor (2th order):
        #------------------------------------------------------------------
        phi_ab = self._get_phi_Emab(kappa)
        #------------------------------------------------------------------
        # Damage tensor (4th order) using product- or sum-type symmetrization:
        #------------------------------------------------------------------
        beta_abcd = self._get_beta_Emabcd(phi_ab)
        #------------------------------------------------------------------
        # Damaged stiffness tensor calculated based on the damage tensor beta4:
        #------------------------------------------------------------------
        D_ijab = einsum(
            '...ijab, abef, ...cdef -> ...ijcd',
            beta_abcd, self.D_abef, beta_abcd
        )

        sig_ab = einsum('...abef,...ef -> ...ab', D_ijab, eps_ab)

        return sig_ab, D_ijab


    '''Number of microplanes - currently fixed for 3D
    '''
    n_mp = Constant(22)

    _alpha_list = Property(depends_on='n_mp')

    @cached_property
    def _get__alpha_list(self):
        return array([np.pi / self.n_mp * (i - 0.5)
                      for i in range(1, self.n_mp + 1)])

    #-----------------------------------------------
    # get the normal vectors of the microplanes
    #-----------------------------------------------

    _MPN = Property(depends_on='n_mp')

    @cached_property
    def _get__MPN(self):
        # microplane normals:

        alpha_list = np.linspace(0, 2 * np.pi, self.n_mp)

        MPN = np.array([[np.cos(alpha), np.sin(alpha)]
                        for alpha in alpha_list])

        return MPN

    #-------------------------------------
    # get the weights of the microplanes
    #-------------------------------------
    _MPW = Property(depends_on='n_mp')

    @cached_property
    def _get__MPW(self):
        # Note that the values in the array must be multiplied by 6 (cf. [Baz05])!
        # The sum of of the array equals 0.5. (cf. [BazLuz04]))
        # The values are given for an Gaussian integration over the unit
        # hemisphere.
        MPW = np.ones(self.n_mp) / self.n_mp * 2

        return MPW



'''
Created on Dec 3, 2015

@author: rch
'''

from matplotlib.pyplot import xlabel
from traits.api import \
    HasStrictTraits, Dict, Property, Float, \
    WeakRef, DelegatesTo, cached_property, \
    Str, List, Button, Bool
from .viz2d import Viz2D


class Viz2DPlot(Viz2D):
    '''Plot adaptor for the pull-out simulator.
    '''
    xlabel = Str(None)

    label = Property

    def _set_label(self, value):
        if value:
            self.xlabel = value

    @cached_property
    def _get_label(self):
        if self.xlabel:
            return xlabel
        else:
            return 'plot: %s' % self.plot_fn

    plot_fn = Str

    def plot(self, ax, vot, *args, **kw):
        getattr(self.vis2d, self.plot_fn)(ax, vot, *args, **kw)


class Viz2DDict(HasStrictTraits):
    '''On demand constructor of viz2d object, 
    Objects are constructed upon access using the key within  
    the viz2d_classes dictionary.
    '''

    vis2d = WeakRef

    viz2d_classes = DelegatesTo('vis2d')

    _viz2d_objects = Dict

    def __getitem__(self, key):
        viz2d = self._viz2d_objects.get(key, None)
        if viz2d == None:
            viz2d_class = self.viz2d_classes.get(key, None)
            if viz2d_class == None:
                raise KeyError('No vizualization class with key %s' % key)
            viz2d = viz2d_class(name=key, vis2d=self.vis2d)
            self._viz2d_objects[key] = viz2d
        return viz2d

    def __delitem__(self, key):
        del self._viz2d_objects[self.__keytransform__(key)]

    def __iter__(self):
        return iter(self._viz2d_objects)

    def __len__(self):
        return len(self._viz2d_objects)

    def __keytransform__(self, key):
        return key

    def __repr__(self):
        return repr(self._viz2d_objects)


class Vis2D(HasStrictTraits):
    '''Each state and operator object can be associated with 
    several visualization objects with a shortened class name Viz3D. 
    In order to introduce a n independent class subsystem into 
    the class structure, objects supporting visualization inherit 
    from Visual3D which introduces a dictionary viz3d objects.
    '''

    def setup(self):
        pass

    sim = WeakRef
    '''Root of the simulation to extract the data
    '''

    vot = Float(0.0, time_change=True)
    '''Visual object time
    '''

    viz2d_classes = Dict
    '''Visualization classes applicable to this object. 
    '''

    viz2d_class_names = Property(List(Str),
                                 depends_on='viz2d_classes')
    '''Keys of the viz2d classes
    '''
    @cached_property
    def _get_viz2d_class_names(self):
        return list(self.viz2d_classes.keys())

    selected_viz2d_class = Str

    def _selected_viz2d_class_default(self):
        if len(self.viz2d_class_names) > 0:
            return self.viz2d_class_names[0]
        else:
            return ''

    add_selected_viz2d = Button(label='Add plot viz2d')

    def _add_selected_viz2d_fired(self):
        viz2d_class_name = self.selected_viz2d_class
        self.add_viz2d(viz2d_class_name, '<unnamed>')

    def add_viz2d(self, class_name, name, **kw):
        if name == '':
            name = class_name
        viz2d_class = self.viz2d_classes[class_name]
        viz2d = viz2d_class(name=name, vis2d=self, **kw)
        self.viz2d.append(viz2d)
        if hasattr(self, 'ui') and self.ui:
            self.ui.viz_sheet.viz2d_list.append(viz2d)

    viz2d = List(Viz2D)

    def plt(self, name, label=None):
        return Viz2DPlot(plot_fn=name, label=label, vis2d=self)





from ibvpy.tmodel.mats2D.mats2D_cmdm.mats2D_cmdm import MATS2DMicroplaneDamage
from ibvpy.tmodel.mats2D.mats2D_elastic.vmats2D_elastic import MATS2DElastic
from ibvpy.tmodel.mats2D.mats2D_sdamage.vmats2D_sdamage import MATS2DScalarDamage
from ibvpy.tmodel.matsXD.matsXD_explore import MATSXDExplore
from ibvpy.util.traits.either_type import \
   EitherType


class MATS2DExplore(MATSXDExplore):
    '''
    Simulate the loading histories of a material point in 2D space.
    '''

    mats_eval = EitherType(klasses=[MATS2DElastic,
                                    MATS2DMicroplaneDamage,
                                    MATS2DScalarDamage ])

    def _mats_eval_default(self):
        return MATS2DElastic()




from .sim.sim_base import Simulator
from .sim.hist import Hist
from .sim.i_hist import IHist
from .sim.i_tmodel import ITModel
from .sim.i_simulator import ISimulator
from .sim.i_tloop import ITLoop
from .sim.i_tstep import ITStep
from .sim.i_xmodel import IXModel
from .sim.tmodel import TModel
from .sim.tline import TLine
from .sim.tloop import TLoop
from .sim.tstep import TStep
from .sim.tstep_bc import TStepBC
from .xmodel.xdomain_fe_grid import XDomainFEGrid
from .xmodel.xdomain_interface1d import XDomainFEInterface1D
from .xmodel.xdomain_lattice import XDomainLattice
from .xmodel.xdomain_point import XDomainSinglePoint
# Boundary condition classes
from .bcond.bc_dof import BCDof
from .bcond.bc_dofgroup import BCDofGroup
from .bcond.bc_slice import BCSlice
from .bcond import BCSliceI
from .bcond.bc_slice import BCSlice as BCSliceE
# Boundary condition classes
from .fets.fets import FETSEval
from .fets.i_fets import IFETSEval
from .fets import FETS2D4Q, FETS3D8H
# Material model classes
from .tmodel import MATSEval, IMATSEval
from .tmodel import MATS1D5BondSlipMultiLinear, MATS1D5D, \
    MATS1D5BondSlipD, MATS1D5BondSlipEP, MATS1D5BondSlipTriLinear
from .tmodel.mats3D.mats3D_sdamage.vmats3D_sdamage import MATS3DScalarDamage
from .tmodel.mats3D.mats3D_elastic.vmats3D_elastic import MATS3DElastic
from .tmodel.mats2D.mats2D_sdamage.vmats2D_sdamage import MATS2DScalarDamage
from .tmodel.mats2D import MATS2DMplDamageEEQ
# FEGrid classes
from .mesh.fe_domain import FEDomain
from .mesh.fe_grid import FEGrid
from .mesh.fe_grid_idx_slice import FEGridIdxSlice
from .mesh.fe_grid_ls_slice import FEGridLevelSetSlice
from .mesh.fe_refinement_grid import FERefinementGrid, FERefinementGrid as FEPatchedGrid
# Time function classes for material models and boundary conditions
from .tfunction import TFCyclicNonsymmetricIncreasing, TimeFunction, TFSelector, \
    TFMonotonic, TFCyclicSymmetricIncreasing, TFCyclicSymmetricConstant, \
    TFCyclicNonsymmetricIncreasing, TFCyclicNonsymmetricConstant, \
    TFCyclicSin, TFBilinear




from traits.api \
    import HasTraits, Array, \
    on_trait_change, Instance, WeakRef, Int, \
    Interface, provides

#------------------------------------------------------------------------
# Source of the data for array view
#------------------------------------------------------------------------
class ICellArraySource(Interface):

    '''Object representing a structured 1,2,3-dimensional cell grid.
    '''
    pass

#------------------------------------------------------------------------
# Cell view interface
#------------------------------------------------------------------------


class ICellView(Interface):

    '''Interface of the general cell view.
    '''

    def set_cell_traits(self):
        '''Adapt the view to the newly set cell_idx.
        '''
        raise NotImplementedError

    def redraw(self):
        '''Redraw the graphical representation of the cell 
        in the mayavi pipeline.
        '''
        raise NotImplementedError

#------------------------------------------------------------------------
# Default implementation of the cell view
#------------------------------------------------------------------------


@provides(ICellView)
class CellView(HasTraits):

    '''Get the element numbers.
    '''

    cell_idx = Int(-1)

    cell_grid = WeakRef(ICellArraySource)

    def set_cell(self, cell_idx):
        '''Method to be overloaded by subclasses. The subclass 
        can fetch the data required from the cell_grid 
        '''
        self.cell_idx = cell_idx
        self.set_cell_traits()

    def set_cell_traits(self):
        '''Specialize this function to fetch the cell data from 
        the array source.
        '''
        pass

    def redraw(self):
        '''No plotting defined by default'''
        pass

class CellArray(HasTraits):

    data = Array

    cell_view = Instance(ICellView)

    def _cell_view_default(self):
        return CellView()

    current_row = Int(-1)

    @on_trait_change('current_row')
    def redraw(self):
        self.cell_view.redraw()

    @on_trait_change('current_row')
    def _display_current_row(self):
        if self.current_row != -1:
            self.cell_view.set_cell(self.current_row)





from scipy.linalg import \
    inv
from traits.api import \
    Array, Float, \
    Instance, Int
from ibvpy.fets.fets_eval import FETSEval
from ibvpy.tmodel.mats_eval import MATSEval
import numpy as np


#------------------------------------------------------------------------------
# FETS2D9Q - 9 nodes isoparametric quadrilateral (2D, quadratic, Lagrange family)
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
# Element Information:
#------------------------------------------------------------------------------
#
# Here an isoparametric element formulation is applied.
# The implemented shape functions are derived in femple
# based on the following ordering of the nodes of the
# parent element:
#
#    _node_coord_map_dof = Array( Float, (9,2),
#                                 [[ -1.,-1. ],
#                                  [  1.,-1. ],
#                                  [  1., 1. ],
#                                  [ -1., 1. ],
#                                  [  0.,-1. ],
#                                  [  1., 0. ],
#                                  [  0., 1. ],
#                                  [ -1., 0. ],
#                                  [  0., 0. ]])
#
#    _node_coord_map_geo = Array( Float, (9,2),
#                                 [[ -1.,-1. ],
#                                  [  1.,-1. ],
#                                  [  1., 1. ],
#                                  [ -1., 1. ],
#                                  [  0.,-1. ],
#                                  [  1., 0. ],
#                                  [  0., 1. ],
#                                  [ -1., 0. ],
#                                  [  0., 0. ]])
#
#------------------------------------------------------------------------------
class FETS2D9Q(FETSEval):
    debug_on = True

    mats_eval = Instance(MATSEval)

    # Dimensional mapping
    dim_slice = slice(0, 2)

    n_e_dofs = Int(9 * 2)
    t = Float(1.0, label='thickness')
    E = Float(1.0, label="Young's modulus")
    nu = Float(0., label="Poisson's ratio")

    # Integration parameters
    #
    ngp_r = 3
    ngp_s = 3

    dof_r = Array(value=[[-1., -1.],
                         [1., -1.],
                         [1., 1.],
                         [-1., 1.],
                         [0., -1.],
                         [1., 0.],
                         [0., 1.],
                         [-1., 0.],
                         [0., 0.]])
    geo_r = Array(value=[[-1., -1.],
                         [1., -1.],
                         [1., 1.],
                         [-1., 1.],
                         [0., -1.],
                         [1., 0.],
                         [0., 1.],
                         [-1., 0.],
                         [0., 0.]])
    #
    vtk_cell_types = 'QuadraticQuad'
    vtk_r = Array(value=[[-1., -1.],
                         [1., -1.],
                         [1., 1.],
                         [-1., 1.],
                         [0., -1.],
                         [1., 0.],
                         [0., 1.],
                         [-1., 0.],
                         [0., 0.]])
    vtk_cells = [[0, 1, 2, 3, 4, 5, 6, 7, 8]]

    n_nodal_dofs = Int(2)

    #---------------------------------------------------------------------
    # Method required to represent the element geometry
    #---------------------------------------------------------------------
    def get_N_geo_mtx(self, r_pnt):
        '''
        Return the value of shape functions for the specified local coordinate r
        '''
        N_geo_mtx = np.zeros((1, 9), dtype='float64')
        N_geo_mtx[0, 0] = (
            r_pnt[0] * r_pnt[1] * (-1 + r_pnt[1]) * (-1 + r_pnt[0])) / 4.0
        N_geo_mtx[0, 1] = (
            r_pnt[0] * r_pnt[1] * (-1 + r_pnt[1]) * (1 + r_pnt[0])) / 4.0
        N_geo_mtx[0, 2] = (
            r_pnt[0] * r_pnt[1] * (1 + r_pnt[1]) * (1 + r_pnt[0])) / 4.0
        N_geo_mtx[0, 3] = (
            r_pnt[0] * r_pnt[1] * (1 + r_pnt[1]) * (-1 + r_pnt[0])) / 4.0
        N_geo_mtx[
            0, 4] = - (r_pnt[1] * (-1 + r_pnt[0]) * (1 + r_pnt[0]) * (-1 + r_pnt[1])) / 2.0
        N_geo_mtx[
            0, 5] = - (r_pnt[0] * (-1 + r_pnt[1]) * (1 + r_pnt[1]) * (1 + r_pnt[0])) / 2.0
        N_geo_mtx[
            0, 6] = - (r_pnt[1] * (-1 + r_pnt[0]) * (1 + r_pnt[0]) * (1 + r_pnt[1])) / 2.0
        N_geo_mtx[
            0, 7] = - (r_pnt[0] * (-1 + r_pnt[1]) * (1 + r_pnt[1]) * (-1 + r_pnt[0])) / 2.0
        N_geo_mtx[0, 8] = (-1 + r_pnt[1]) * \
            (1 + r_pnt[1]) * (-1 + r_pnt[0]) * (1 + r_pnt[0])
        return N_geo_mtx

    def get_dNr_geo_mtx(self, r_pnt):
        '''
        Return the matrix of shape function derivatives.
        Used for the construction of the Jacobi matrix.
        '''
        dNr_geo_mtx = np.zeros((2, 9), dtype='float64')
        dNr_geo_mtx[0, 0] = (r_pnt[1] * (-1 + r_pnt[1]) * (-1 + r_pnt[0])
                             ) / 4.0 + (r_pnt[0] * r_pnt[1] * (-1 + r_pnt[1])) / 4.0
        dNr_geo_mtx[0, 1] = (r_pnt[1] * (-1 + r_pnt[1]) * (1 + r_pnt[0])
                             ) / 4.0 + (r_pnt[0] * r_pnt[1] * (-1 + r_pnt[1])) / 4.0
        dNr_geo_mtx[0, 2] = (r_pnt[1] * (1 + r_pnt[1]) * (1 + r_pnt[0])
                             ) / 4.0 + (r_pnt[0] * r_pnt[1] * (1 + r_pnt[1])) / 4.0
        dNr_geo_mtx[0, 3] = (r_pnt[1] * (1 + r_pnt[1]) * (-1 + r_pnt[0])
                             ) / 4.0 + (r_pnt[0] * r_pnt[1] * (1 + r_pnt[1])) / 4.0
        dNr_geo_mtx[0, 4] = - (r_pnt[1] * (-1 + r_pnt[1]) * (1 + r_pnt[0])) / 0.2e1 - (
            r_pnt[1] * (-1 + r_pnt[1]) * (-1 + r_pnt[0])) / 2.0
        dNr_geo_mtx[0, 5] = - ((-1 + r_pnt[1]) * (1 + r_pnt[1]) * (1 + r_pnt[0])) / 2.0 - (
            r_pnt[0] * (-1 + r_pnt[1]) * (1 + r_pnt[1])) / 2.0
        dNr_geo_mtx[0, 6] = - (r_pnt[1] * (1 + r_pnt[1]) * (1 + r_pnt[0])
                               ) / 2.0 - (r_pnt[1] * (1 + r_pnt[1]) * (-1 + r_pnt[0])) / 2.0
        dNr_geo_mtx[0, 7] = - ((-1 + r_pnt[1]) * (1 + r_pnt[1]) * (-1 + r_pnt[0])) / 2.0 - (
            r_pnt[0] * (-1 + r_pnt[1]) * (1 + r_pnt[1])) / 2.0
        dNr_geo_mtx[0, 8] = (-1 + r_pnt[1]) * (1 + r_pnt[1]) * (1 +
                                                                r_pnt[0]) + (-1 + r_pnt[1]) * (1 + r_pnt[1]) * (-1 + r_pnt[0])
        dNr_geo_mtx[1, 0] = (r_pnt[0] * (-1 + r_pnt[1]) * (-1 + r_pnt[0])
                             ) / 4.0 + (r_pnt[0] * r_pnt[1] * (-1 + r_pnt[0])) / 4.0
        dNr_geo_mtx[1, 1] = (r_pnt[0] * (-1 + r_pnt[1]) * (1 + r_pnt[0])
                             ) / 4.0 + (r_pnt[0] * r_pnt[1] * (1 + r_pnt[0])) / 4.0
        dNr_geo_mtx[1, 2] = (r_pnt[0] * (1 + r_pnt[1]) * (1 + r_pnt[0])
                             ) / 4.0 + (r_pnt[0] * r_pnt[1] * (1 + r_pnt[0])) / 4.0
        dNr_geo_mtx[1, 3] = (r_pnt[0] * (1 + r_pnt[1]) * (-1 + r_pnt[0])
                             ) / 4.0 + (r_pnt[0] * r_pnt[1] * (-1 + r_pnt[0])) / 4.0
        dNr_geo_mtx[1, 4] = - ((-1 + r_pnt[0]) * (1 + r_pnt[0]) * (-1 + r_pnt[1])) / 2.0 - (
            r_pnt[1] * (-1 + r_pnt[0]) * (1 + r_pnt[0])) / 2.0
        dNr_geo_mtx[1, 5] = - (r_pnt[0] * (1 + r_pnt[1]) * (1 + r_pnt[0])
                               ) / 2.0 - (r_pnt[0] * (-1 + r_pnt[1]) * (1 + r_pnt[0])) / 2.0
        dNr_geo_mtx[1, 6] = - ((-1 + r_pnt[0]) * (1 + r_pnt[0]) * (1 + r_pnt[1])) / 2.0 - (
            r_pnt[1] * (-1 + r_pnt[0]) * (1 + r_pnt[0])) / 2.0
        dNr_geo_mtx[1, 7] = - (r_pnt[0] * (1 + r_pnt[1]) * (-1 + r_pnt[0])) / 2.0 - (
            r_pnt[0] * (-1 + r_pnt[1]) * (-1 + r_pnt[0])) / 2.0
        dNr_geo_mtx[1, 8] = (-1 + r_pnt[0]) * (1 + r_pnt[0]) * (1 +
                                                                r_pnt[1]) + (-1 + r_pnt[0]) * (1 + r_pnt[0]) * (-1 + r_pnt[1])
        return dNr_geo_mtx

    #---------------------------------------------------------------------
    # Method delivering the shape functions for the field variables and their derivatives
    #---------------------------------------------------------------------
    def get_N_mtx(self, r_pnt):
        '''
        Returns the matrix of the shape functions used for the field approximation
        containing zero entries. The number of rows corresponds to the number of nodal
        dofs. The matrix is evaluated for the specified local coordinate r_pnt.
        '''
        N = self.get_N_geo_mtx(r_pnt)
        I_mtx = np.identity(self.n_nodal_dofs, float)
        N_mtx_list = [I_mtx * N[0, i] for i in range(0, N.shape[1])]
        N_mtx = np.hstack(N_mtx_list)
        return N_mtx

    def get_dNr_mtx(self, r_pnt):
        '''
        Return the derivatives of the shape functions used for the field approximation
        '''
        dNr_mtx = self.get_dNr_geo_mtx(r_pnt)
        return dNr_mtx

    def get_B_mtx(self, r_pnt, X_mtx):
        J_mtx = self.get_J_mtx(r_pnt, X_mtx)
        dNr_mtx = self.get_dNr_mtx(r_pnt)
        dNx_mtx = np.dot(inv(J_mtx), dNr_mtx)
        Bx_mtx = np.zeros((3, 18), dtype='float64')
        for i in range(0, 9):
            Bx_mtx[0, i * 2] = dNx_mtx[0, i]
            Bx_mtx[1, i * 2 + 1] = dNx_mtx[1, i]
            Bx_mtx[2, i * 2] = dNx_mtx[1, i]
            Bx_mtx[2, i * 2 + 1] = dNx_mtx[0, i]
        return Bx_mtx

#----------------------- example --------------------

if __name__ == '__main__':
    from ibvpy.api import \
        TStepper as TS, RTDofGraph, RTraceDomainListField, TLoop, \
        TLine, BCDofGroup

    #from lib.tmodel.mats2D.mats_cmdm2D.mats_mdm2d import MACMDM
    #from ibvpy.tmodel.mats2D.mats2D_sdamage.mats2D_sdamage import MATS2DScalarDamage
    from ibvpy.tmodel.mats2D.mats2D_elastic.mats2D_elastic import MATS2DElastic

    fets_eval = FETS2D9Q(mats_eval=MATS2DElastic())
    #fets_eval = FETS2D9Q(mats_eval = MACMDM())

    from ibvpy.mesh.fe_grid import FEGrid

    # Discretization
    domain = FEGrid(coord_max=(3., 3., 0.),
                    shape=(4, 4),
                    fets_eval=fets_eval)

    right_dof = 2
    ts = TS(sdomain=domain,
            # conversion to list (square brackets) is only necessary for slicing of
            # single dofs, e.g "get_left_dofs()[0,1]"
            bcond_list=[BCDofGroup(var='u', value=0., dims=[0],
                                   get_dof_method=domain.get_left_dofs),
                        BCDofGroup(var='u', value=0., dims=[1],
                                   get_dof_method=domain.get_bottom_left_dofs),
                        BCDofGroup(var='u', value=0.002, dims=[0],
                                   get_dof_method=domain.get_right_dofs)],
            rtrace_list=[RTDofGraph(name='Fi,right over u_right (iteration)',
                                     var_y='F_int', idx_y=right_dof,
                                     var_x='U_k', idx_x=right_dof),
                         #                         RTraceDomainListField(name = 'Stress' ,
                         #                         var = 'sig_app', idx = 0,
                         #                         record_on = 'update'),
                         RTraceDomainListField(name='Displacement',
                                               var='u', idx=0),
                         #                             RTraceDomainListField(name = 'N0' ,
                         #                                          var = 'N_mtx', idx = 0,
                         # record_on = 'update')

                         ]
            )

    # Add the time-loop control
    #
    tl = TLoop(tstepper=ts,
               DT=0.5,
               tline=TLine(min=0.0,  max=1.0))

    tl.eval()
    # Put the whole stuff into the simulation-framework to map the
    # individual pieces of definition into the user interface.
    #
    from ibvpy.plugins.ibvpy_app import IBVPyApp
    app = IBVPyApp(ibv_resource=tl)
    app.main()




from math import floor
from mayavi.core.source import Source
from numpy import zeros, mgrid, c_, array, arange, \
    ones, random
from traits.api import \
    HasTraits, Float, Int, Array, Property, cached_property, \
    Tuple, List, Str, on_trait_change, Button, Delegate, \
    Instance
from functools import reduce

# tvtk related imports
#
#---------------------------------------------------------------------
# C L A S S  GridPoint
#---------------------------------------------------------------------
class GridPoint(HasTraits):
    '''
    Editable point specification to be used for orientation 
    within the grid.
    '''
    _coords = List(Float, [0., 0., 0.], enter_set=True, auto_set=False)

    x = Property(Float, enter_set=True, auto_set=False)
    y = Property(Float, enter_set=True, auto_set=False)
    z = Property(Float, enter_set=True, auto_set=False)

    def _get_x(self):
        '''get the value of x'''
        return self._coords[0]

    def _set_x(self, value):
        '''set the value of x'''
        self._coords[0] = value

    def _get_y(self):
        '''get the value of y'''
        return self._coords[1]

    def _set_y(self, value):
        '''set the value of y'''
        self._coords[1] = value

    def _get_z(self):
        '''get the value of z'''
        return self._coords[2]

    def _set_z(self, value):
        '''set the value of z'''
        self._coords[2] = value

    def __getitem__(self, idx):
        '''Delegate access to _coords'''
        return self._coords[idx]

    def __iter__(self):
        '''Delegate iteration to _coords'''
        return iter(self._coords)

#---------------------------------------------------------------------
# C L A S S  GeoNDGrid
#---------------------------------------------------------------------


class GeoNDGrid(Source):
    '''
    Specification and representation of an nD-grid.

    GridND
    '''
    # The name of our scalar array.
    scalar_name = Str('scalar')

    # map of coordinate labels to the indices
    _dim_map = {'x': 0, 'y': 1, 'z': 2}

    # currently active dimensions
    active_dims = List(Str, ['x', 'y'])

    # Bottom left corner
    x_mins = Instance(GridPoint, label='Corner 1')

    def _x_mins_default(self):
        '''Bottom left corner'''
        return GridPoint()
    # Upper right corner
    x_maxs = Instance(GridPoint, label='Corner 2')

    def _x_maxs_default(self):
        '''Upper right corner'''
        return GridPoint(x=1, y=1, z=1)

    # indices of the currently active dimensions
    dim_indices = Property(Array(int), depends_on='active_dims')

    @cached_property
    def _get_dim_indices(self):
        ''' Get active indices '''
        return array([self._dim_map[dim_ix]
                      for dim_ix
                      in self.active_dims], dtype='int_')

    # number of currently active dimensions
    n_dims = Property(Int, depends_on='active_dims')

    @cached_property
    def _get_n_dims(self):
        '''Number of currently active dimensions'''
        return len(self.active_dims)

    # number of elements in each direction
    # @todo: rename to n_faces
    shape = Tuple(int, int, int, label='Elements')

    def _shape_default(self):
        '''Number of elements in each direction'''
        return (1, 0, 0)

    n_act_nodes = Property(Array, depends_on='shape, active_dims')

    @cached_property
    def _get_n_act_nodes(self):
        '''Number of active nodes respecting the active_dim'''
        act_idx = ones((3, ), int)
        shape = array(list(self.shape), dtype=int)
        act_idx[self.dim_indices] += shape[self.dim_indices]
        return act_idx

    # total number of nodes of the system grid
    n_nodes = Property(Int, depends_on='shape, active_dims')

    @cached_property
    def _get_n_nodes(self):
        '''Number of nodes used for the geometry approximation'''
        return reduce(lambda x, y: x * y, self.n_act_nodes)

    enum_nodes = Property(Array, depends_on='shape,active_dims')

    @cached_property
    def _get_enum_nodes(self):
        '''
        Returns an array of element numbers respecting the grid structure
        (the nodes are numbered first in x-direction, then in y-direction and
        last in z-direction)
        '''
        return arange(self.n_nodes).reshape(tuple(self.n_act_nodes))

    grid = Property(Array,
                    depends_on='shape,active_dims,x_mins.+,x_maxs.+')

    @cached_property
    def _get_grid(self):
        '''
        slice(start,stop,step) with step of type 'complex' leads to that number of divisions 
        in that direction including 'stop' (see numpy: 'mgrid')
        '''
        slices = [slice(x_min, x_max, complex(0, n_n))
                  for x_min, x_max, n_n
                  in zip(self.x_mins, self.x_maxs, self.n_act_nodes)]
        return mgrid[tuple(slices)]

    #-------------------------------------------------------------------------
    # Visualization pipelines
    #-------------------------------------------------------------------------
#     mvp_mgrid_geo = Trait(MVPolyData)
#
#     def _mvp_mgrid_geo_default(self):
#         return MVPolyData(name='Mesh geomeetry',
#                           points=self._get_points,
#                           lines=self._get_lines,
#                           polys=self._get_faces,
#                           scalars=self._get_random_scalars
#                           )
#
#     mvp_mgrid_labels = Trait(MVPointLabels)
#
#     def _mvp_mgrid_labels_default(self):
#         return MVPointLabels(name='Mesh numbers',
#                              points=self._get_points,
#                              scalars=self._get_random_scalars,
#                              vectors=self._get_points)

    changed = Button('Draw')

    @on_trait_change('changed')
    def redraw(self):
        '''
        '''
        self.mvp_mgrid_geo.redraw()
        self.mvp_mgrid_labels.redraw('label_scalars')

    def _get_points(self):
        '''
        Reshape the grid into a column.
        '''
        return c_[tuple([self.grid[i].flatten() for i in range(3)])]

    def _get_n_lines(self):
        '''
        Get the number of lines.
        '''
        act_idx = ones((3, ), int)
        act_idx[self.dim_indices] += self.shape[self.dim_indices]
        return reduce(lambda x, y: x * y, act_idx)

    def _get_lines(self):
        '''
        Only return data if n_dims = 1
        '''
        if self.n_dims != 1:
            return array([], int)
        #
        # Get the list of all base nodes
        #
        tidx = ones((3,), dtype='int_')
        tidx[self.dim_indices] = -1
        slices = tuple([slice(0, idx) for idx in tidx])
        base_node_list = self.enum_nodes[slices].flatten()
        #
        # Get the node map within the line
        #
        ijk_arr = zeros((3, 2), dtype=int)
        ijk_arr[self.dim_indices[0]] = [0, 1]
        offsets = self.enum_nodes[ijk_arr[0], ijk_arr[1], ijk_arr[2]]
        #
        # Setup and fill the array with line connectivities
        #
        n_lines = self._get_n_lines()
        lines = zeros((n_lines, 2), dtype='int_')
        for n_idx, base_node in enumerate(base_node_list):
            lines[n_idx, :] = offsets + base_node
        return lines

    def _get_n_faces(self):
        '''Return the number of faces.

        The number is determined by putting 1 into inactive dimensions and 
        shape into the active dimensions. 
        '''
        act_idx = ones((3, ), int)
        shape = array(self.shape, dtype=int)
        act_idx[self.dim_indices] = shape[self.dim_indices]
        return reduce(lambda x, y: x * y, act_idx)

    def _get_faces(self):
        '''
        Only return data of n_dims = 2.
        '''
        if self.n_dims != 2:
            return array([], int)
        #
        # get the slices extracting all corner nodes with
        # the smallest node number within the element
        #
        tidx = ones((3,), dtype='int_')
        tidx[self.dim_indices] = -1
        slices = tuple([slice(0, idx) for idx in tidx])
        base_node_list = self.enum_nodes[slices].flatten()
        #
        # get the node map within the face
        #
        ijk_arr = zeros((3, 4), dtype=int)
        ijk_arr[self.dim_indices[0]] = [0, 0, 1, 1]
        ijk_arr[self.dim_indices[1]] = [0, 1, 1, 0]
        offsets = self.enum_nodes[ijk_arr[0], ijk_arr[1], ijk_arr[2]]
        #
        # setup and fill the array with line connectivities
        #
        n_faces = self._get_n_faces()
        faces = zeros((n_faces, 4), dtype='int_')
        for n_idx, base_node in enumerate(base_node_list):
            faces[n_idx, :] = offsets + base_node
        return faces

    def _get_volumes(self):
        '''
        Only return data if ndims = 3
        '''
        if self.n_dims != 3:
            return array([], int)

        tidx = ones((3,), dtype='int_')
        tidx[self.dim_indices] = -1
        slices = tuple([slice(0, idx) for idx in tidx])

        en = self.enum_nodes
        offsets = array([en[0, 0, 0], en[0, 1, 0], en[1, 1, 0], en[1, 0, 0],
                         en[0, 0, 1], en[0, 1, 1], en[1, 1, 1], en[1, 0, 1]], dtype='int_')
        base_node_list = self.enum_nodes[slices].flatten()

        n_faces = self._get_n_faces()
        faces = zeros((n_faces, 8), dtype='int_')
        for n in base_node_list:
            faces[n, :] = offsets + n

        return faces

    # Identifiers
    var = Str('dummy')
    idx = Int(0)

    def _get_random_scalars(self):
        return random.weibull(1, size=self.n_nodes)




'''
Created on Mar 2, 2017

@author: rch
'''

from builtins import len
import os
import tempfile
from threading import Thread

from matplotlib.figure import \
    Figure
from mayavi.core.ui.api import \
    MayaviScene, SceneEditor, MlabSceneModel
from pyface.api import GUI
from ibvpy.view.reporter import ROutputSection
from ibvpy.sim.i_hist import IHist
from traits.api import \
    Str, Instance,  Event, Enum, \
    Tuple, List, Range, Int, Float, \
    Property, cached_property, \
    on_trait_change, Bool, Button, Directory, \
    HasStrictTraits, WeakRef
from ibvpy.util.traits.editors import \
    MPLFigureEditor
from ibvpy.view.plot2d.viz2d import Viz2D
from ibvpy.view.plot3d.viz3d import Viz3D

import matplotlib.pyplot as plt
import numpy as np
import traits.api as tr


class RunThread(Thread):
    '''Time loop thread responsible.
    '''

    def __init__(self, vs, vot, *args, **kw):
        super(RunThread, self).__init__(*args, **kw)
        self.daemon = True
        self.vs = vs
        self.vot = vot

    def run(self):
        # print 'STARTING THREAD'
        GUI.invoke_later(self.vs.update_pipeline, self.vot)
        # print 'THREAD ENDED'


class Viz2DAdapter(TabularAdapter):
    # List of (Column labels, Column ID).
    columns = [
        ('Label',    'label'),
        ('Visible', 'visible'),
    ]


# The tabular editor works in conjunction with an adapter class, derived from
# the 'players' trait table editor:
viz2d_list_editor = TableEditor(
    sortable=False,
    configurable=False,
    auto_size=False,
    selected='selected_viz2d',
    click='viz2d_list_editor_clicked',
    columns=[CheckboxColumn(name='visible',  label='visible',
                            width=0.12),
             ObjectColumn(name='name', editable=False, width=0.24,
                          horizontal_alignment='left'),
             ])

# The tabular editor works in conjunction with an adapter class, derived from
# the 'players' trait table editor:
pp_list_editor = TableEditor(
    sortable=False,
    configurable=False,
    auto_size=False,
    selected='selected_pp',
    columns=[ObjectColumn(name='name', editable=False, width=0.24,
                          horizontal_alignment='left'),
             ])

# The tabular editor works in conjunction with an adapter class, derived from
# the 'players' trait table editor:
viz3d_list_editor = TableEditor(
    sortable=False,
    configurable=False,
    auto_size=False,
    selected='selected_viz3d',
    #    click='viz3d_list_editor_clicked',
    columns=[CheckboxColumn(name='visible',  label='visible',
                            width=0.12),
             ObjectColumn(name='name', editable=False, width=0.24,
                          horizontal_alignment='left'),
             ])


class AnimationDialog(HasStrictTraits):
    '''Animation parameters for a diagram.
    '''
    sheet = WeakRef

    export_path = Directory
    status_message = Str('')

    animate_from = Float(0.0, auto_set=False, enter_set=True)
    animate_to = Float(1.0, auto_set=False, enter_set=True)
    animate_steps = Int(30, auto_set=False, enter_set=True)

    animate_button = Button(label='Animate selected diagram')

    def _animate_button_fired(self):

        if self.export_path == '':
            dir_ = tempfile.mkdtemp()
        else:
            dir_ = self.export_path
        name = self.sheet.selected_viz2d.name
        path = os.path.join(dir_, name)

        if os.path.exists(path):
            self.status_message = 'overwriting animation %s' % name
        else:
            os.makedirs(path)

        for i, vot in enumerate(np.linspace(self.animate_from,
                                            self.animate_to,
                                            self.animate_steps)):
            fname = os.path.join(path, 'step%03d.jpg' % i)
            self.sheet.selected_viz2d.savefig_animate(
                vot, fname,
                (self.sheet.fig_width,
                 self.sheet.fig_height)
            )
        self.status_message = 'animation stored in %s' % path

class PlotPerspective(HasStrictTraits):
    name = Str('<unnamed>')
    viz2d_list = List(Viz2D, input=True)
    positions = List([], input=True)
    twinx = List([], input=True)
    twiny = List([], input=True)
    data_changed = Event

    figure = Instance(Figure)

    tight_layout = Bool(True)

    def _figure_default(self):
        figure = Figure(facecolor='white')
        figure.set_tight_layout(self.tight_layout)
        return figure

    viz2d_axes = Property

    @cached_property
    def _get_viz2d_axes(self):
        return {viz2d: self.figure.add_subplot(loc)
                for viz2d, loc in zip(self.viz2d_list,
                                      self.positions)}

    twinx_axes = Property

    @cached_property
    def _get_twinx_axes(self):
        return {viz_2: self.viz2d_axes[viz_1].twinx()
                for viz_1, viz_2, _ in self.twinx}

    twiny_axes = Property

    @cached_property
    def _get_twiny_axes(self):
        return {viz_2: self.viz2d_axes[viz_1].twiny()
                for viz_1, viz_2, _ in self.twiny}

    axes = Property

    @cached_property
    def _get_axes(self):
        ad = {}
        ad.update(self.viz2d_axes)
        ad.update(self.twinx_axes)
        ad.update(self.twiny_axes)
        return ad

    def clear(self):
        for viz2d, ax in self.viz2d_axes.items():
            ax.clear()
            viz2d.reset(ax)
        for viz2d, ax in self.twinx_axes.items():
            ax.clear()
            viz2d.reset(ax)
        for viz2d, ax in self.twiny_axes.items():
            ax.clear()
            viz2d.reset(ax)

    def replot(self, vot):
        for viz2d, ax in self.axes.items():
            ax.clear()
            viz2d.clear()
            viz2d.plot(ax, vot)
        self.data_changed = True

    def align_xaxis(self):
        for v1, v2, alignx in self.twiny:
            if alignx:
                ax1 = self.viz2d_axes[v1]
                ax2 = self.twiny_axes[v2]
                self._align_xaxis(ax1, ax2)

    def _align_xaxis(self, ax1, ax2):
        """Align zeros of the two axes, zooming them out by same ratio"""
        axes = (ax1, ax2)
        extrema = [ax.get_xlim() for ax in axes]
        tops = [extr[1] / (extr[1] - extr[0]) for extr in extrema]
        # Ensure that plots (intervals) are ordered bottom to top:
        if tops[0] > tops[1]:
            axes, extrema, tops = [list(reversed(l))
                                   for l in (axes, extrema, tops)]

        # How much would the plot overflow if we kept current zoom levels?
        tot_span = tops[1] + 1 - tops[0]

        b_new_t = extrema[0][0] + tot_span * (extrema[0][1] - extrema[0][0])
        t_new_b = extrema[1][1] - tot_span * (extrema[1][1] - extrema[1][0])
        axes[0].set_xlim(extrema[0][0], b_new_t)
        axes[1].set_xlim(t_new_b, extrema[1][1])

class BMCSVizSheet(ROutputSection):
    '''Vieualization sheet
    - controls the time displayed
    - contains several vizualization adapters.
    This class could be called BMCSTV - for watching the time
    dependent response. It can have several channels - in 2D and 3D
    '''

    def __init__(self, *args, **kw):
        super(BMCSVizSheet, self).__init__(*args, **kw)
        self.on_trait_change(self.viz2d_list_items_changed,
                             'viz2d_list_items')

    name = Str

    hist = Instance(IHist)

    min = Float(0.0)
    '''Simulation start is always 0.0
    '''
    max = Float(1.0)
    '''Upper range limit of the current simulator.
    This range is determined by the the time-loop range
    of the model. 
    '''
    vot = Float

    def _vot_default(self):
        return self.min

    def _vot_changed(self):
        if self.hist:
            self.hist.vot = self.vot

    vot_slider = Range(low='min', high='max', step=0.01,
                       enter_set=True, auto_set=False)
    '''Time line controlling the current state of the simulation.
    this value is synchronized with the control time of the
    time loop setting the tline. The vot_max = tline.max.
    The value of vot follows the value of tline.val in monitoring mode.
    By default, the monitoring mode is active with vot = tline.value.
    When sliding to a value vot < tline.value, the browser mode is activated.
    When sliding into the range vot > tline.value the monitoring mode
    is reactivated. 
    '''

    def _vot_slider_default(self):
        return 0.0

    mode = Enum('monitor', 'browse')

    def _mode_changed(self):
        if self.mode == 'browse':
            self.offline = False

    time = Float(0.0)

    def time_range_changed(self, max_):
        self.max = max_

    def time_changed(self, time):
        self.time = time
        if self.mode == 'monitor':
            self.vot = time
            self.vot_slider = time

    def _vot_slider_changed(self):
        if self.mode == 'browse':
            if self.vot_slider >= self.time:
                self.mode = 'monitor'
                self.vot_slider = self.time
                self.vot = self.time
            else:
                self.vot = self.vot_slider
        elif self.mode == 'monitor':
            if self.vot_slider < self.time:
                self.mode = 'browse'
                self.vot = self.vot_slider
            else:
                self.vot_slider = self.time
                self.vot = self.time

    offline = Bool(True)
    '''If the sheet is offline, the plot refresh is inactive.
    The sheet starts in offline mode and is activated once the signal
    run_started has been received. Upon run_finished the 
    the sheet goes directly into the offline mode again.
    
    If the user switches to browser mode, the vizsheet gets online 
    and reploting is activated.
    '''

    running = Bool(False)

    def run_started(self):
        self.running = True
        self.offline = False
        for pp in self.pp_list:
            pp.clear()
        self.mode = 'monitor'
        if self.reference_viz2d:
            ax = self.reference_axes
            ax.clear()
            self.reference_viz2d.reset(ax)

    def run_finished(self):
        self.skipped_steps = self.monitor_chunk_size
        # self.update_pipeline(1.0)
        self.replot()
        self.running = False
        self.offline = True

    monitor_chunk_size = Int(10, label='Monitor each # steps')

    skipped_steps = Int(1)

    @on_trait_change('vot,n_cols')
    def replot(self):
        if self.offline:
            return
        if self.running and self.mode == 'monitor' and \
                self.skipped_steps < (self.monitor_chunk_size - 1):
            self.skipped_steps += 1
            return
        for pp in self.pp_list:
            pp.replot(self.vot)
#         for viz2d, ax in self.axes.items():
#             ax.clear()
#             viz2d.clear()
#             viz2d.plot(ax, self.vot)
#         if self.selected_pp:
#             self.selected_pp.align_xaxis()
        if self.reference_viz2d:
            ax = self.reference_axes
            ax.clear()
            self.reference_viz2d.clear()
            self.reference_viz2d.plot(ax, self.vot)
        self.data_changed = True
        self.skipped_steps = 0
        if self.mode == 'browse':
            self.update_pipeline(self.vot)
        else:
            up = RunThread(self, self.vot)
            up.start()

    viz2d_list = List(Viz2D)
    '''List of visualization adaptors for 2D.
    '''
    viz2d_dict = Property

    def _get_viz2d_dict(self):
        return {viz2d.name: viz2d for viz2d in self.viz2d_list}

    viz2d_names = Property
    '''Names to be supplied to the selector of the
    reference graph.
    '''

    def _get_viz2d_names(self):
        return list(self.viz2d_dict.keys())

    viz2d_list_editor_clicked = Tuple
    viz2d_list_changed = Event

    def _viz2d_list_editor_clicked_changed(self, *args, **kw):
        _, column = self.viz2d_list_editor_clicked
        self.offline = False
        self.viz2d_list_changed = True
        if self.plot_mode == 'single':
            if column.name == 'visible':
                self.selected_viz2d.visible = True
                self.plot_mode = 'multiple'
            else:
                self.replot()
        elif self.plot_mode == 'multiple':
            if column.name != 'visible':
                self.plot_mode = 'single'
            else:
                self.replot()

    plot_mode = Enum('multiple', 'single')

    def _plot_mode_changed(self):
        if self.plot_mode == 'single':
            self.replot_selected_viz2d()
        elif self.plot_mode == 'multiple':
            self.replot()

    def replot_selected_viz2d(self):
        for viz2d in self.viz2d_list:
            viz2d.visible = False
        self.selected_viz2d.visible = True
        self.n_cols = 1
        self.viz2d_list_changed = True
        self.replot()

    def viz2d_list_items_changed(self):
        self.replot()

    def get_subrecords(self):
        '''What is this good for?
        '''
        return self.viz2d_list

    export_button = Button(label='Export selected diagram')

    def plot_in_window(self):
        fig = plt.figure(figsize=(self.fig_width, self.fig_height))
        ax = fig.add_subplot(111)
        self.selected_viz2d.plot(ax, self.vot)
        fig.show()

    def _export_button_fired(self, vot=0):
        print('in export button fired')
        Thread(target=self.plot_in_window).start()
        print('thread started')

    fig_width = Float(8.0, auto_set=False, enter_set=True)
    fig_height = Float(5.0, auto_set=False, enter_set=True)

    save_button = Button(label='Save selected diagram')

    animate_button = Button(label='Animate selected diagram')

    def _animate_button_fired(self):
        ad = AnimationDialog(sheet=self)
        ad.edit_traits()
        return
    #=========================================================================
    # Reference figure serving for orientation.
    #=========================================================================
    reference_viz2d_name = Enum('', values="viz2d_names")
    '''Current name of the reference graphs.
    '''

    def _reference_viz2d_name_changed(self):
        self.replot()

    reference_viz2d_cumulate = Bool(False, label='cumulate')
    reference_viz2d = Property(Instance(Viz2D),
                               depends_on='reference_viz2d_name')
    '''Visualization of a graph showing the time context of the
    current visualization state. 
    '''

    def _get_reference_viz2d(self):
        if self.reference_viz2d_name == None:
            if len(self.viz2d_dict):
                return self.viz2d_list[0]
            else:
                return None
        return self.viz2d_dict[self.reference_viz2d_name]

    reference_figure = Instance(Figure)

    def _reference_figure_default(self):
        figure = Figure(facecolor='white')
        figure.set_tight_layout(True)
        return figure

    reference_axes = Property(List,
                              depends_on='reference_viz2d_name')
    '''Derived axes objects reflecting the layout of plot pane
    and the individual. 
    '''
    @cached_property
    def _get_reference_axes(self):
        return self.reference_figure.add_subplot(1, 1, 1)

    selected_viz2d = Instance(Viz2D)

    def _selected_viz2d_changed(self):
        if self.plot_mode == 'single':
            self.replot_selected_viz2d()

    n_cols = Range(low=1, high=3, value=2, label='Number of columns',
                   tooltip='Defines a number of columns within the plot pane',
                   enter_set=True, auto_set=False)

    figure = Instance(Figure)

    tight_layout = Bool(True)

    def _figure_default(self):
        figure = Figure(facecolor='white')
        figure.set_tight_layout(self.tight_layout)
        return figure

    visible_viz2d_list = Property(List,
                                  depends_on='viz2d_list,viz2d_list_items,n_cols,viz2d_list_changed')
    '''Derived axes objects reflecting the layout of plot pane
    and the individual. 
    '''
    @cached_property
    def _get_visible_viz2d_list(self):
        viz_list = []
        for viz2d in self.viz2d_list:
            if viz2d.visible:
                viz_list.append(viz2d)
        return viz_list

    pp_list = List(PlotPerspective)

    selected_pp = Instance(PlotPerspective)

    xaxes = Property(List,
                     depends_on='selected_pp')
    '''Derived axes objects reflecting the layout of plot pane
    and the individual. 
    '''
    @cached_property
    def _get_xaxes(self):
        self.figure.clear()
        if self.selected_pp:
            self.selected_pp.figure = self.figure
            ad = self.selected_pp.axes
        else:
            n_fig = len(self.visible_viz2d_list)
            n_cols = self.n_cols
            n_rows = (n_fig + n_cols - 1) / self.n_cols
            ad = {viz2d: self.figure.add_subplot(n_rows, self.n_cols, i + 1)
                  for i, viz2d in enumerate(self.visible_viz2d_list)}
        return ad

    data_changed = Event

    bgcolor = tr.Tuple(1.0, 1.0, 1.0)
    fgcolor = tr.Tuple(0.0, 0.0, 0.0)

    scene = Instance(MlabSceneModel)

    def _scene_default(self):
        return MlabSceneModel()

    mlab = Property(depends_on='input_change')
    '''Get the mlab handle'''

    def _get_mlab(self):
        return self.scene.mlab

    fig = Property()
    '''Figure for 3D visualization.
    '''
    @cached_property
    def _get_fig(self):
        fig = self.mlab.gcf()
        bgcolor = tuple(self.bgcolor)
        fgcolor = tuple(self.fgcolor)
        self.mlab.figure(fig, fgcolor=fgcolor, bgcolor=bgcolor)
        return fig

    def show(self, *args, **kw):
        '''Render the visualization.
        '''
        self.mlab.show(*args, **kw)

    def add_viz3d(self, viz3d, order=1):
        '''Add a new visualization objectk.'''
        viz3d.ftv = self
        vis3d = viz3d.vis3d
        name = viz3d.name
        label = '%s[%s:%s]-%s' % (name,
                                  str(vis3d.__class__),
                                  str(viz3d.__class__),
                                  vis3d
                                  )
        if label in self.viz3d_dict:
            raise KeyError('viz3d object named %s already registered' % label)
        viz3d.order = order
        self.viz3d_dict[label] = viz3d

    viz3d_dict = tr.Dict(tr.Str, tr.Instance(Viz3D))
    '''Dictionary of visualization objects.
    '''

    viz3d_list = tr.Property

    def _get_viz3d_list(self):
        map_order_viz3d = {}
        for idx, (viz3d) in enumerate(self.viz3d_dict.values()):
            order = viz3d.order
            map_order_viz3d['%5g%5g' % (order, idx)] = viz3d
        return [map_order_viz3d[key] for key in sorted(map_order_viz3d.keys())]

    pipeline_ready = Bool(False)

    def setup_pipeline(self):
        if self.pipeline_ready:
            return
        self.fig
        fig = self.mlab.gcf()
        fig.scene.disable_render = True
        for viz3d in self.viz3d_list:
            viz3d.setup()
        fig.scene.disable_render = False
        self.pipeline_ready = True

    def update_pipeline(self, vot):
        self.setup_pipeline()
        # get the current constrain information
        self.vot = vot
        fig = self.mlab.gcf()
        fig.scene.disable_render = True
        for viz3d in self.viz3d_list:
            viz3d.plot(vot)
        fig.scene.disable_render = False

    selected_viz3d = Instance(Viz3D)

    def _selected_viz3d_changed(self):
        print('selection done')




'''
Created on Feb 11, 2018

@author: rch
'''

import os
import threading

from mayavi import mlab
from mayavi.filters.api import ExtractTensorComponents
from mayavi.modules.api import Surface
from mayavi.sources.vtk_xml_file_reader import VTKXMLFileReader
from tvtk.api import \
    tvtk, write_data
from ibvpy.view.plot3d.viz3d import Vis3D, Viz3D

import numpy as np
import traits.api as tr


class Vis3DSDamage(Vis3D):

    def setup(self, tloop):
        self.new_dir()
        self.tloop = tloop
        fets = tloop.ts.fets
        ts = self.tloop.ts
        DELTA_ab = fets.vtk_expand_operator

        vtk_cell_type = fets.vtk_cell_class().cell_type
        n_c = fets.n_nodal_dofs
        n_i = fets.n_dof_r
        n_E, n_i, n_a = ts.x_Eia.shape
        n_Ei = n_E * n_i
        points = np.einsum(
            'Ia,ab->Ib', ts.x_Eia.reshape(-1, n_c), DELTA_ab)
        U = np.zeros_like(points)
        self.ug = tvtk.UnstructuredGrid(points=points)
        vtk_cells = (np.arange(n_E) * n_i)[:, np.newaxis] + \
            np.array(fets.vtk_cell, dtype=np.int_)[np.newaxis, :]
        self.ug.set_cells(vtk_cell_type, vtk_cells)
        self.update(U, 0)

    file_list = tr.List(tr.Str,
                        desc='a list of files belonging to a time series')

    def update(self, U, t):
        ts = self.tloop.ts
        fets = ts.fets
        omega_field = ts.state_arrays['omega']
        n_c = fets.n_nodal_dofs
        DELTA_ab = fets.vtk_expand_operator
        U_Ia = U.reshape(-1, n_c)
        U_Eia = U_Ia[ts.I_Ei]
        U_vector_field = np.einsum(
            'Ia,ab->Ib', U_Eia.reshape(-1, n_c), DELTA_ab
        )
        self.ug.point_data.vectors = U_vector_field
        self.ug.point_data.vectors.name = 'displacement'
        self.ug.point_data.scalars = omega_field.flatten()
        self.ug.point_data.scalars.name = 'damage'
        fname = 'omega_step_%008.4f' % t
        target_file = os.path.join(
            self.dir, fname.replace('.', '_')
        ) + '.vtu'
        write_data(self.ug, target_file)
        self.add_file(target_file)


class Viz3DSDamage(Viz3D):

    label = tr.Str('damage')
    vis3d = tr.WeakRef

    warp_factor = tr.Float(1.0, auto_set=False, enter_set=True)

    def setup(self):
        m = mlab
        fname = self.vis3d.file_list[0]
        self.d = VTKXMLFileReader()
        self.d.initialize(fname)
        self.src = m.pipeline.add_dataset(self.d)
        self.warp_vector = m.pipeline.warp_vector(self.src)
        self.warp_vector.filter.scale_factor = self.warp_factor
        self.surf = m.pipeline.surface(self.warp_vector)
        lut = self.warp_vector.children[0]
        lut.scalar_lut_manager.set(
            lut_mode='Reds',
            show_scalar_bar=True,
            show_legend=True,
            data_name='damage',
            use_default_range=False,
            data_range=np.array([0, 1], dtype=np.float64)
        )

    def plot(self, vot):
        idx = self.vis3d.tloop.get_time_idx(vot)
        self.d.file_list = self.vis3d.file_list
        self.d.timestep = idx + 1




from math import exp, sin

from traits.api import \
    Enum, Float,  \
    Trait,  Event, \
    Dict
from ibvpy.tmodel.mats1D.mats1D_eval import MATS1DEval
import numpy as np


#---------------------------------------------------------------------------
# Material time-step-evaluator for Scalar-Damage-Model
#---------------------------------------------------------------------------
class MATS1DDamage(MATS1DEval):
    '''
    Scalar Damage Model.
    '''

    E = Float(1.,  # 34e+3,
              modified=True,
              label="E",
              desc="Young's Modulus",
              enter_set=True,
              auto_set=False)

    epsilon_0 = Float(1.,  # 59e-6,
                      modified=True,
                      label="eps_0",
                      desc="Breaking Strain",
                      enter_set=True,
                      auto_set=False)

    epsilon_f = Float(1.,  # 191e-6,
                      modified=True,
                      label="eps_f",
                      desc="Shape Factor",
                      enter_set=True,
                      auto_set=False)

    stiffness = Enum("secant", "algorithmic",
                     modified=True)

    # This event can be used by the clients to trigger an action upon
    # the completed reconfiguration of the material model
    #
    changed = Event

    #--------------------------------------------------------------------------
    # View specification
    #--------------------------------------------------------------------------

    #-------------------------------------------------------------------------
    # Setup for computation within a supplied spatial context
    #-------------------------------------------------------------------------

    def get_state_array_size(self):
        '''
        Give back the nuber of floats to be saved
        @param sctx:spatial context
        '''
        return 2

    def new_cntl_var(self):
        return np.zeros(1, np.float64)

    def new_resp_var(self):
        return np.zeros(1, np.float64)

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------

    def get_corr_pred(self, sctx, eps_app_eng, d_eps, tn, tn1, eps_avg=None):
        '''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''
        if eps_avg == None:
            eps_avg = eps_app_eng

        E = self.E
        D_el = np.array([E])

        if sctx.update_state_on:

            kappa_n = sctx.mats_state_array[0]
            kappa_k = sctx.mats_state_array[1]
            sctx.mats_state_array[0] = kappa_k

        kappa_k, omega = self._get_state_variables(sctx, eps_avg)
        sctx.mats_state_array[1] = kappa_k

        if self.stiffness == "algorithmic":
            D_e_dam = np.array([self._get_alg_stiffness(sctx, eps_app_eng,
                                                        kappa_k,
                                                        omega)])
        else:
            D_e_dam = np.array([(1 - omega) * D_el])

        sigma = np.dot(np.array([(1 - omega) * D_el]), eps_app_eng)

        # print the stress you just computed and the value of the apparent E

        return sigma, D_e_dam

    #--------------------------------------------------------------------------
    # Subsidiary methods realizing configurable features
    #--------------------------------------------------------------------------

    def _get_state_variables(self, sctx, eps):

        kappa_n, kappa_k = sctx.mats_state_array

        kappa_k = max(abs(eps), kappa_n)

        omega = self._get_omega(sctx, kappa_k)

        return kappa_k, omega

    def _get_omega(self, sctx, kappa):
        epsilon_0 = self.epsilon_0
        epsilon_f = self.epsilon_f
        if kappa >= epsilon_0:
            return 1. - epsilon_0 / kappa * exp(-1 * (kappa - epsilon_0) / epsilon_f)
        else:
            return 0.

    def _get_alg_stiffness(self, sctx, eps_app_eng, e_max, omega):
        E = self.E
        D_el = np.array([E])
        epsilon_0 = self.epsilon_0
        epsilon_f = self.epsilon_f
        dodk = (epsilon_0 / (e_max * e_max) * exp(-(e_max - epsilon_0) / epsilon_f) +
                epsilon_0 / e_max / epsilon_f * exp(-(e_max - epsilon_0) / epsilon_f))
        D_alg = (1 - omega) * D_el - D_el * eps_app_eng * dodk
        return D_alg

    #--------------------------------------------------------------------------
    # Response trace evaluators
    #--------------------------------------------------------------------------

    def get_omega(self, sctx, eps_app_eng, eps_avg=None):
        if eps_avg == None:
            eps_avg = eps_app_eng
        return self._get_omega(sctx, eps_avg)

    # Declare and fill-in the rte_dict - it is used by the clients to
    # assemble all the available time-steppers.
    #
    rte_dict = Trait(Dict)

    def _rte_dict_default(self):
        return {'sig_app': self.get_sig_app,
                'eps_app': self.get_eps_app,
                'omega': self.get_omega}

    #-------------------------------------------------------------------------
    # List of response tracers to be constructed within the mats_explorer
    #-------------------------------------------------------------------------
    def _get_explorer_rtrace_list(self):
        '''Return the list of relevant tracers to be used in mats_explorer.
        '''
        return []

    def _get_explorer_config(self):
        from ibvpy.api import TLine, RTDofGraph, BCDof
        ec = super(MATS1DDamage, self)._get_explorer_config()
        ec['mats_eval'] = MATS1DDamage(E=1.0, epsilon_0=1.0, epsilon_f=5)
        ec['bcond_list'] = [BCDof(var='u',
                                  dof=0, value=1.7,
                                  time_function=lambda t: (1 + 0.1 * t) * sin(t))]
        ec['tline'] = TLine(step=0.1, max=10)
        ec['rtrace_list'] = [
            RTDofGraph(name='strain - stress',
                       var_x='eps_app', idx_x=0,
                       var_y='sig_app', idx_y=0,
                       record_on='update'),
            RTDofGraph(name='time - damage',
                       var_x='time', idx_x=0,
                       var_y='omega', idx_y=0,
                       record_on='update')
        ]
        return ec



from matplotlib.figure import \
    Figure
from scipy import interpolate as ip
from traits.api import Array, Float, Event, \
    ToolbarButton, on_trait_change, \
    Property, cached_property, Enum, Instance, Bool
from ibvpy.view.ui import BMCSLeafNode

import matplotlib.pyplot as plt
import numpy as np


class MFnLineArray(BMCSLeafNode):

    # Public Traits
    xdata = Array(float, value=[0.0, 1.0])

    def _xdata_default(self):
        '''
        convenience default - when xdata not defined created automatically as
        an array of integers with the same shape as ydata
        '''
        return np.arange(self.ydata.shape[0])

    ydata = Array(float, value=[0.0, 1.0])

    def __init__(self, *args, **kw):
        super(MFnLineArray, self).__init__(*args, **kw)
        self.replot()

    extrapolate = Enum('constant', 'exception', 'diff', 'zero')
    '''
    Vectorized interpolation using scipy.interpolate
    '''

    def values(self, x, k=1):
        '''
        vectorized interpolation, k is the spline order, default set to 1 (linear)
        '''
        tck = ip.splrep(self.xdata, self.ydata, s=0, k=k)

        x = np.array([x]).flatten()

        if self.extrapolate == 'diff':
            values = ip.splev(x, tck, der=0)
        elif self.extrapolate == 'exception':
            if x.all() < self.xdata[0] and x.all() > self.xdata[-1]:
                values = values = ip.splev(x, tck, der=0)
            else:
                raise ValueError('value(s) outside interpolation range')
        elif self.extrapolate == 'constant':
            values = ip.splev(x, tck, der=0)
            values[x < self.xdata[0]] = self.ydata[0]
            values[x > self.xdata[-1]] = self.ydata[-1]
        elif self.extrapolate == 'zero':
            values = ip.splev(x, tck, der=0)
            values[x < self.xdata[0]] = 0.0
            values[x > self.xdata[-1]] = 0.0
        return values

    def __call__(self, x):
        return self.values(x)

    yrange = Property
    '''Get min max values on the vertical axis
    '''

    def _get_yrange(self):
        return np.min(self.ydata), np.max(self.ydata)

    xrange = Property
    '''Get min max values on the vertical axis
    '''

    def _get_xrange(self):
        return np.min(self.xdata), np.max(self.xdata)

    data_changed = Event

    figure = Instance(Figure)

    def _figure_default(self):
        figure = Figure(facecolor='white')
        return figure

    def diff(self, x, k=1, der=1):
        '''
        vectorized interpolation, der is the nth derivative, default set to 1;
        k is the spline order of the data inetrpolation, default set to 1 (linear)
        '''
        xdata = np.sort(np.hstack((self.xdata, x)))
        idx = np.argwhere(np.diff(xdata) == 0).flatten()
        xdata = np.delete(xdata, idx)
        tck = ip.splrep(xdata, self.values(xdata, k=k), s=0, k=k)
        return ip.splev(x, tck, der=der)

    dump_button = ToolbarButton('Print data',
                                style='toolbar')

    @on_trait_change('dump_button')
    def print_data(self, event=None):
        print('x = ', repr(self.xdata))
        print('y = ', repr(self.ydata))

    integ = Property(Float(), depends_on='ydata')

    @cached_property
    def _get_integ(self):
        _xdata = self.xdata
        _ydata = self.ydata
        # integral under the stress strain curve
        return np.trapz(_ydata, _xdata)

    def clear(self):
        self.xdata = np.array([])
        self.ydata = np.array([])

    def plot(self, axes, *args, **kw):
        self.mpl_plot(axes, *args, **kw)

    def mpl_plot(self, axes, *args, **kw):
        '''plot within matplotlib window'''
        axes.plot(self.xdata, self.ydata, *args, **kw)

    def mpl_plot_diff(self, axes, *args, **kw):
        '''plot within matplotlib window'''
        ax_dx = axes.twinx()
        x = np.linspace(self.xdata[0], self.xdata[-1],
                        np.size(self.xdata) * 20.0)
        y_dx = self.diff(x, k=1, der=1)
        ax_dx.plot(x, y_dx, *args + ('-',), **kw)

    plot_diff = Bool(False)

    def replot(self):
        self.figure.clf()
        ax = self.figure.add_subplot(111)
        self.mpl_plot(ax)
        if self.plot_diff:
            self.mpl_plot_diff(ax, color='orange')
        self.data_changed = True

    def savefig(self, fname):
        fig = plt.figure()
        ax = fig.add_subplot(111)
        self.mpl_plot(ax)
        self.mpl_plot_diff(ax, color='orange')
        fig.savefig(fname)




from numpy import \
    ix_, array, int_, dot, newaxis, float64, copy, repeat
from traits.api import Array, Bool, Enum, Float, HasTraits, \
    Int, Trait, Enum, \
    Callable, List, \
    Button, \
    provides

from .i_bcond import \
    IBCond
# from ibvpy.view.plot3d.mayavi_util.pipelines import \
#     MVPointLabels

from .bc_dof import BCDof

@provides(IBCond)
class BCDofGroup(HasTraits):

    '''
    Implements the IBC functionality for a constrained dof.
    '''
    var = Enum('u', 'f')

    get_dof_method = Callable

    get_link_dof_method = Callable  # optional callable to deliver
    # dofs for linked pairing
    dof_numbers = Array(int)
    dof_X = Array(Float)

    bcdof_list = List(BCDof)

    # List of dofs that determine the value of the current dof
    #
    # If this list is empty, then the current dof is
    # prescribed. Otherwise, the dof value is given by the
    # linear combination of DOFs in the list (see the example below)
    #
    link_dofs = List(Int)

    # Coefficients of the linear combination of DOFs specified in the
    # above list.
    #
    link_coeffs = List(Float)

    dims = List(Int)

    link_dims = List(Int)

    value = Float

    # TODO - adapt the definition
    time_function = Callable

    def _time_function_default(self):
        return lambda t: t

    def is_essential(self):
        return self.var == 'u'

    def is_linked(self):
        return self.get_link_dof_method != None

    def is_constrained(self):
        '''
        Return true if a DOF is either explicitly prescribed or it depends on other DOFS.
        '''
        return self.is_essential() or self.is_linked()

    def is_natural(self):
        return self.var == 'f'

    def get_dofs(self):
        return list(self.dof_numbers.flatten())

    def setup(self, sctx):
        '''
        Locate the spatial context.
        '''
        dof_numbers, self.dof_X = self.get_dof_method()
        self.dof_numbers = dof_numbers[:, tuple(self.dims)]

        if self.get_link_dof_method:
            link_dofs, linked_dof_X = self.get_link_dof_method()
            # TODO: test if the shape is the same as dof_numbers
            if len(self.link_dims) == 0:
                self.link_dims = self.dims
            else:
                if len(self.dims) != len(self.link_dims):
                    raise IndexError('incompatible dim specification (%d != %d'
                                     % (len(self.dims), len(self.link_dims)))

            link_dofs_arr = link_dofs[:, tuple(self.link_dims)]

            # slice the dof_numbers for the proper direction
            #
            self.bcdof_list = [BCDof(var=self.var, dof=dof,
                                     value=self.value,
                                     link_dofs=[ldof],
                                     link_coeffs=self.link_coeffs,
                                     time_function=self.time_function)
                               for dof, ldof in zip(self.dof_numbers.flatten(), link_dofs_arr.flatten())]
        else:
            self.bcdof_list = [BCDof(var=self.var, dof=dof,
                                     value=self.value,
                                     link_dofs=self.link_dofs,
                                     link_coeffs=self.link_coeffs,
                                     time_function=self.time_function)
                               for dof in self.dof_numbers.flatten()]

        return

    def apply_essential(self, K):

        for bcond in self.bcdof_list:
            bcond.apply_essential(K)

    def apply(self, step_flag, sctx, K, R, t_n, t_n1):

        for bcond in self.bcdof_list:
            bcond.apply(step_flag, sctx, K, R, t_n, t_n1)

    # register the pipelines for plotting labels and geometry
    #
    # mvp_dofs = Trait(MVPointLabels)
    #
    # def _mvp_dofs_default(self):
    #     return MVPointLabels(name='Boundary condition',
    #                          points=self._get_mvpoints,
    #                          vectors=self._get_labels,
    #                          color=(0.0, 0.0, 0.882353))

    def _get_mvpoints(self):
        # blow up
        print('dof_X', self.dof_X)
        return array(self.dof_X, dtype='float64')

    def _get_labels(self):
        # blow up
        n_points = self.dof_numbers.shape[0]
        dofs = repeat(-1., n_points * 3).reshape(n_points, 3)
        dofs[:, tuple(self.dims)] = self.dof_numbers
        print('BC - DOFS', dofs)
        return dofs

    redraw_button = Button('Redraw')

    def _redraw_button_fired(self):
        self.mvp_dofs.redraw(label_mode='label_vectors')



'''
Created on 14. 4. 2014

@author: rch
'''

from ibvpy.view.reporter import RInputRecord
from traits.api import \
    HasStrictTraits, Str, List, WeakRef, \
    Property, on_trait_change, Event, Any
import bmcs_utils.api as bu

itags = dict(
    TIME=True,
    ALG=True,
    GEO=True,
    MESH=True,
    MAT=True,
    FE=True,
    CS=True,
    BC=True
)

itags_str = ','.join(['%s' % tag for tag in itags])


class BMCSNodeBase(bu.Model):

    node_name = Str('<unnamed>')

    ipw_view = bu.View()

    def update_plot(self, axes):
        pass

    ui = WeakRef

    def set_traits_with_metadata(self, value, **metadata):
        traits_names = self.trait_names(**metadata)
        for tname in traits_names:
            setattr(self, tname, value)

    parent = WeakRef

    root = Property
    '''Root node of tree node hierarchy
    '''

    def _get_root(self):
        if self.parent:
            return self.parent.root
        return self


class BMCSListeningTreeNodeMixIn(HasStrictTraits):

    @on_trait_change('+TIME')
    def _TIME_change(self):
        if self.parent:
            # print 'TIME change'
            self.root.TIME = True

    @on_trait_change('+ALG')
    def _ALG_change(self):
        if self.parent:
            # print 'ALG change'
            self.root.ALG = True

    @on_trait_change('+GEO')
    def _GEO_change(self):
        if self.parent:
            # print 'GEO change'
            self.root.GEO = True

    @on_trait_change('+MESH')
    def _MESH_change(self):
        if self.parent:
            # print 'MESH change'
            self.root.MESH = True

    @on_trait_change('+MAT')
    def _MAT_change(self):
        if self.parent:
            # print 'MAT change'
            self.root.MAT = True

    @on_trait_change('+FE')
    def _FE_change(self):
        if self.parent:
            # print 'FE change'
            self.root.FE = True

    @on_trait_change('+CS')
    def _CS_change(self):
        if self.parent:
            # print 'CS change'
            self.root.CS = True

    @on_trait_change('+BC')
    def _BC_change(self):
        if self.parent:
            # print 'BC change'
            self.root.BC = True


class BMCSLeafNodeMixIn(HasStrictTraits):
    '''Base class of all model classes that can appear in a tree view.
    '''

    def set_parents_recursively(self):
        return

    def set_ui_recursively(self, ui):
        self.ui = ui


class BMCSTreeNodeMixIn(HasStrictTraits):
    '''Base class of all model classes that can appear in a tree view.
    '''
    tree_node_list = List([])

    def _tree_node_list_items_changed(self):
        self.set_parents_recursively()

    def _tree_node_list_changed(self):
        self.set_parents_recursively()

    def set_parents_recursively(self):
        for n in self.tree_node_list:
            if n is None:
                continue
            n.parent = self
            n.set_parents_recursively()

    def set_ui_recursively(self, ui):
        self.ui = ui
        for node in self.tree_node_list:
            node.set_ui_recursively(ui)

    def append_node(self, node):
        '''Add a new subnode to the current node.
        Inform the tree view to select the new node within the view.
        '''
        node.set_ui_recursively(self.ui)
        self.tree_node_list.append(node)

    def set_traits_with_metadata(self, value, **metadata):
        super(BMCSTreeNodeMixIn, self).set_traits_with_metadata(
            value, **metadata)
        for node in self.tree_node_list:
            node.set_traits_with_metadata(value, **metadata)


class BMCSLeafNode(BMCSNodeBase,
                   RInputRecord,
                   BMCSLeafNodeMixIn,
                   BMCSListeningTreeNodeMixIn):
    itags = itags


class BMCSTreeNode(BMCSNodeBase,
                   RInputRecord,
                   BMCSTreeNodeMixIn,
                   BMCSListeningTreeNodeMixIn):
    itags = itags


class BMCSRootNode(BMCSNodeBase,
                   RInputRecord,
                   BMCSTreeNodeMixIn,
                   BMCSListeningTreeNodeMixIn):
    '''Root node for the model hierarchy.

    Types of generic change events within a numerical simulation

    time discretization
    algorithm parameters
    geometry (dimensions, scaling, geometrical transformation)
    spatial discretization (spatial meshing / decomposition parameters)
    spatial approximation (finite elements - what happens within an element)
    cross section (thickness, area, perimeter)
    boundary conditions

    '''

    itags = itags

    TIME = Event
    ALG = Event
    GEO = Event
    MESH = Event
    MAT = Event
    FE = Event
    CS = Event
    BC = Event

    def __init__(self, *args, **kw):
        super(BMCSRootNode, self).__init__(*args, **kw)
        self.set_parents_recursively()

    @on_trait_change('+TIME')
    def _TIME_change(self):
        # print 'TIME change'
        self.root.TIME = True

    @on_trait_change('+ALG')
    def _ALG_change(self):
        # print 'ALG change'
        self.root.ALG = True

    @on_trait_change('+GEO')
    def _GEO_change(self):
        # print 'GEO change'
        self.root.GEO = True

    @on_trait_change('+MESH')
    def _MESH_change(self):
        # print 'MESH change'
        self.root.MESH = True

    @on_trait_change('+MAT')
    def _MAT_change(self):
        # print 'MAT change'
        self.root.MAT = True

    @on_trait_change('+FE')
    def _FE_change(self):
        # print 'FE change'
        self.root.FE = True

    @on_trait_change('+CS')
    def _CS_change(self):
        # print 'CS change'
        self.root.CS = True

    @on_trait_change('+BC')
    def _BC_change(self):
        # print 'BC change'
        self.root.BC = True



'''
Created on Jul 9, 2017

@author: rch

Using the reporter
==================
Reporter uses the report items

For single calculation
----------------------
put the inputs and outputs as separate report items

For parametric studies
----------------------
put the input as the input report item
how to mark the parameters that are varied?
can it be done by setting the metadata dynamically with value range?
then the reporter might specify the range of values included.

The p-study driver might gather the parameter values from
the traits and define a schedule of the study. 

A cumulative viz_adapter might be used to gather the results
of the simulation in the desired form.

In other 
An extra list of
viz_adapters might be defined to plot the values from 
the calculation additively.
'''

from io import StringIO
import os
from shutil import copyfile
import subprocess
import sys
import tempfile

from traits.api import \
    HasStrictTraits, Str, List, Property, \
    cached_property, Instance, Constant

from .report_item import RInputRecord, ROutputSection


class ReportStudy(HasStrictTraits):

    name = Str

    title = Str('<no title>')
    desc = Str('''<no description>''')

    input = Instance(RInputRecord)

    output = Instance(ROutputSection)

    preamble_subfile = Constant(r'''\documentclass[main.tex]{subfiles}
\begin{document}
''')

    postable_subfile = Constant(r'''\end{document}
    ''')

    def write_subfile_tex(self, main_file, examples_dir, itags):
        study_name = self.name
        rel_study_path = os.path.join('examples', study_name)
        rdir = os.path.join(examples_dir, study_name)
        if not os.path.exists(rdir):
            os.mkdir(rdir)
        subfile_tex_name = study_name + '.tex'
        rfile_tex = os.path.join(rdir, subfile_tex_name)

        with open(rfile_tex, 'w') as subfile:
            subfile.write(self.preamble_subfile)
            subfile.write(r'''\begin{bmcsex}{%s}{%s}
\noindent %s \\
\begin{center}
            ''' % (self.title, study_name, self.desc))
            self.write_tex_input(subfile, rdir,
                                 rel_study_path, itags)
            self.write_tex_output(subfile, rdir,
                                  rel_study_path, itags)
            subfile.write(r'''\end{center}
            ''')
            subfile.write(r'''\end{bmcsex}
''')
            subfile.write(self.postable_subfile)

        main_file.write(r'''\subfile{%s}
        ''' % os.path.join(rdir, study_name))

    name = Property

    def _get_name(self):
        return self.input.name

    def write_tex_input(self, subfile, rdir,
                        rel_study_path, itags):
        self.input.write_tex_table(subfile, rdir,
                                   rel_study_path, itags)

    def write_tex_output(self, subfile, rdir,
                         rel_study_path, itags):
        if self.output:
            self.output.write_records(subfile, rdir,
                                      rel_study_path, itags)

    def yield_rinput_traits(self):
        model = self.input
        return model.yield_rinput_section('', self.itags)


class Reporter(HasStrictTraits):

    INPUTTAGS = List(['MAT', 'GEO', 'CS', 'ALG', 'BC', 'MESH'])

    itags = Property(depends_on='INPUTTAGS')

    @cached_property
    def _get_itags(self):
        return {tag: True for tag in self.INPUTTAGS}

    studies = List(ReportStudy)

    preamble = Constant(r'''
% !TeX document-id = {c563f3ab-d450-45cf-a935-b20492a6f8a6}
% !TeX program = pdflatex
% !BIB program = biber
% !TeX encoding = UTF-8
% !TeX spellcheck = en_GB
% Requires:
% texlive 2016: Ubuntu 14.04 and 16.04 PPA: ppa:jonathonf/texlive
%  pdflatex --enable-write18, biber, makeglossary
%  perl, ghostscript
%
\documentclass{lib/scidoc}
%
\input{lib/packages}
\input{lib/example}
\input{lib/bmcsex}
%\input{shortcuts/defs}
%\input{bib/bibliographies}

\title{Brittle-Matrix Composite Structures}
\author{Rostislav Chudoba}
%
\begin{document}
%
%\maketitle

\pagenumbering{arabic}
%
% TABLE OF CONTENTS
%\microtypesetup{protrusion=false}  % disable protrusion
%\pdfbookmark[0]{Contents}{toc}
%\tableofcontents % prints Table of Contents
% LISTS AND GLOSSARIES
%\listoffigures
%  \clearpage
%  \listoftables
%  \clearpage
%  \printglossary[title = Nomenclature,  %
%           type = main,    %
%          ]
%  \glsresetall
%  \clearpage
\microtypesetup{protrusion=true}  % enables protrusion
\clearpage
% 
% CONTENT
%
%
''')

    postamble = Constant(r'''
\end{document}
        ''')

    report_name = Str('example_report')

    report_dir = Property()

    @cached_property
    def _get_report_dir(self):
        tmpdir = tempfile.mkdtemp()
        report_dir = os.path.join(tmpdir, self.report_name)
        os.mkdir(report_dir)
        return report_dir

    example_dirname = Str('examples')

    example_dir = Property

    @cached_property
    def _get_example_dir(self):
        exa_dir = os.path.join(self.report_dir, self.example_dirname)
        os.mkdir(exa_dir)
        return exa_dir

    tex_file = Property(depends_on='report_name')

    @cached_property
    def _get_tex_file(self):
        return os.path.join(self.report_dir, self.report_name + '.tex')

    pdf_file = Property(depends_on='report_name')

    @cached_property
    def _get_pdf_file(self):
        return os.path.join(self.report_dir, self.report_name + '.pdf')

    def copy_lib_files(self):
        this_dir, this_filename = os.path.split(__file__)
        tex_source_lib = os.path.join(this_dir, 'texfiles')
#        tex_source_lib = os.path.join(os.sep, 'home', 'rch', 'bmcs_preamble')
        target_dir = os.path.join(self.report_dir, 'lib')
        os.mkdir(target_dir)
        files = ['packages.tex', 'bmcsex.tex', 'example.tex', 'scidoc.cls']
        for f in files:
            copyfile(os.path.join(tex_source_lib, f),
                     os.path.join(target_dir, f))

    def write(self):
        self.copy_lib_files()
        rfile_io = StringIO()
        for study in self.studies:
            study.write_subfile_tex(rfile_io, self.example_dir, self.itags)
            rfile_str = rfile_io.getvalue()

        with open(self.tex_file, "w") as f:
            f.write(self.preamble)
            f.write(rfile_str)
            f.write(self.postamble)

    def show_tex(self):
        with open(self.tex_file, 'r') as f:
            texsource = f.read()
            print(texsource)

    def run_pdflatex(self):
        cmd = ['pdflatex', '-interaction', 'nonstopmode', self.tex_file]
        proc = subprocess.Popen(cmd, cwd=self.report_dir)
        proc.communicate()

        retcode = proc.returncode
        if not retcode == 0:
            raise ValueError('Error {} executing command: {}'.format(
                retcode, ' '.join(cmd)))

    def show_pdf(self):
        print('platform', sys.platform)
        if sys.platform == 'linux':
            subprocess.call(["xdg-open", self.pdf_file])
        else:
            os.startfile(self.pdf_file)



'''
Created on Feb 11, 2018

@author: rch
'''

import copy
import os
import tempfile

from tvtk.api import tvtk
from ibvpy.view.plot2d import Vis2D, Viz2D

import numpy as np
import traits.api as tr


class Vis2DField(Vis2D):

    model = tr.DelegatesTo('sim')

    x_file = tr.File
    file_list = tr.List(tr.File)

    var = tr.Str('<unnamed>')

    dir = tr.Directory

    def new_dir(self):
        self.dir = tempfile.mkdtemp()

    def setup(self):
        self.new_dir()
        # make a loop over the DomainState
        fe_domain = self.sim.tstep.fe_domain
        domain = fe_domain[2]
        xdomain = domain.xmodel
        r_Eia = np.einsum(
            'Eira,Eia->Eir',
            xdomain.T_Emra[..., :xdomain.x_Eia.shape[-1]], xdomain.x_Eia
        )

        file_name = 'slice_x_%s' % (self.var,)
        target_file = os.path.join(
            self.dir, file_name.replace('.', '_') + '.npy'
        )
        #print('r', r_Eia[..., :-1])
        np.save(target_file, r_Eia[..., :-1])
        self.x_file = target_file

    def get_x_Eir(self):
        return np.load(self.x_file)

    def update(self):
        ts = self.sim.tstep
        fe_domain = self.sim.tstep.fe_domain
        domain = fe_domain[2]
        xdomain = domain.xmodel
        U = ts.U_k
        t = ts.t_n1
        s_Emr = xdomain.map_U_to_field(U)

        var_function = domain.tmodel.var_dict.get(self.var, None)
        if var_function == None:
            raise ValueError('no such variable' % self.var)

        state_k = copy.deepcopy(domain.state_n)
        var_k = var_function(s_Emr, ts.t_n1, **state_k)

        target_file = self.filename(t)

        #np.save(target_file, s_Emr)
        np.save(target_file, var_k)
        self.file_list.append(target_file)

    def filename(self, t):
        file_name = 'slice_%s_step_%008.4f' % (self.var, t)
        target_file = os.path.join(
            self.dir, file_name.replace('.', '_')
        ) + '.npy'
        return target_file


class Viz2DField(Viz2D):

    def setup(self):
        print('Viz2DField:seetup')
        pass

    adaptive_y_range = tr.Bool(True)
    initial_plot = tr.Bool(True)
    y_min = tr.Float(0)
    y_max = tr.Float(0)

    def plot(self, ax, vot, *args, **kw):
        idx = self.vis2d.sim.hist.get_time_idx(vot)
        if len(self.vis2d.file_list) <= idx:
            return
        file_name = self.vis2d.file_list[idx]
        X_Eir = self.vis2d.get_x_Eir()
        x_sorted = X_Eir[..., 0].flatten()
        var_Eir = np.load(file_name)
        slip_sorted = var_Eir.flatten()
        ax.plot(x_sorted, slip_sorted, linewidth=2, color='blue')
        ax.fill_between(x_sorted, slip_sorted,
                        facecolor='blue', alpha=0.2)
#        ax.plot([0, L], [0, 0], color='black')
        ax.set_ylabel(self.vis2d.var)
        ax.set_xlabel('bond length')
        ymin, ymax = np.min(slip_sorted), np.max(slip_sorted)

        if self.adaptive_y_range:
            if self.initial_plot:
                self.y_max = ymax
                self.y_min = ymin
                self.initial_plot = False
                return
        self.y_max = max(ymax, self.y_max)
        self.y_min = min(ymin, self.y_min)
        ax.set_ylim(ymin=self.y_min, ymax=self.y_max)



'''
Created on Apr 23, 2010

@author: alexander
'''

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from traits.trait_base \
    import user_name_for
from bmcs_utils.api import \
    Item


#-------------------------------------------------------------------------
#  'Item' class:
#-------------------------------------------------------------------------
class Item(Item):
    """An element in a Traits-based user interface. 
    The 'get_label' method is overloaded in order to add
    the unit string defined in the trait's metadata as 'unit'
    to the label string.
    """

    def get_label(self, ui):
        """ Gets the label to use for a specified Item.
        """
        # Return 'None' if the Item is a separator or spacer:
        if self.is_spacer():
            return None

        label = self.label
        if label != '':
            return label
        name = self.name
        object_ = eval(self.object_, globals(), ui.context)
        trait = object_.base_trait(name)

        # --------------------------------
        # append metadata 'unit' to the label string if it exists:
        dict_ = trait.__dict__
        if 'unit' in dict_:
            unit_str = ' [' + dict_['unit'] + ']'
        else:
            unit_str = ' '
        label = user_name_for(name) + unit_str
        #---------------------------------

        tlabel = trait.label
        if tlabel is None:
            return label

        if isinstance(tlabel, str):
            if tlabel[0:3] == '...':
                return label + tlabel[3:]
            if tlabel[-3:] == '...':
                return tlabel[:-3] + label
            if self.label != '':
                return self.label
            return tlabel

        return tlabel(object, name, label)
